from __future__ import annotations
from rbceq2.core_logic.constants import AlleleState

from rbceq2.core_logic.alleles import Allele, BloodGroup, Pair
from rbceq2.core_logic.utils import Zygosity

from icecream import ic

def flatten_alleles(pairs: list[Pair]) -> set[Allele]:
    """Flatten the pairs into a set of alleles.

    Args:
        pairs (list[Pair]): A list of Pair objects, where each Pair is an
            iterable containing Allele objects.

    Returns:
        set[Allele]: A set containing all Allele objects from the given pairs.
    """
    return {allele for pair in pairs for allele in pair}


def all_hom(variant_pool: dict[str, str], current_allele: Allele) -> bool:
    """if > 1 het vars in an alleles defiing variant set, this shouldnt apply
    unless phased, which is handled elsewhere - TODO ensure its handled elsewhere"""

    return all(
        variant_pool.get(allele_var) == Zygosity.HOM
        for allele_var in current_allele.defining_variants
    )


def identify_unphased(bg: BloodGroup, alleles: list[Allele]) -> list[Allele]:
    '''
    logic to check if an alleles varaints are in phase or not'''
    def get_var_phase_info(current_var):
        return bg.variant_pool_phase[current_var], bg.variant_pool_phase_set[current_var]
    
    def variant_not_phased(current_phase):
        return '/' in current_phase
    
    def variant_not_HET(current_var):
        return bg.variant_pool.get(current_var) != Zygosity.HET
    
    
    to_remove = []
    for allele in alleles:
        allele_added = False
        for variant in allele.defining_variants:
            if variant_not_HET(variant):
                continue
            phase, phase_set = get_var_phase_info(variant)
            if variant_not_phased(phase):
                continue
            for variant2 in allele.defining_variants:
                if variant == variant2:
                    continue
                if variant_not_HET(variant2):
                    continue
                phase2, phase_set2 = get_var_phase_info(variant2)
                if variant_not_phased(phase2):
                    continue
                if phase_set == phase_set2 and phase != phase2:
                    to_remove.append(allele)
                    allele_added = True
                    break
            if allele_added:
                break
        if allele_added:
            continue
    return to_remove


def proceed(bg: BloodGroup, allele_state: AlleleState) -> bool:
    '''Some filter functions iterate over AlleleStates
    ie for allele_state in [AlleleState.NORMAL, AlleleState.CO]:
    
    if CO, we only want to process knops and then only if there are 
    instances of co_exsisting alleles'''

    if allele_state != AlleleState.CO:
        return True
    if bg.type != "KN" or bg.alleles[allele_state] is None:
        return False
    return True

def check_var(bg: BloodGroup, pair: Pair, allele_state: AlleleState, variant: str) -> int:
    '''checks that HET vars are used'''
    allele1_vars_plus_het_var = set(pair.allele1.defining_variants) | {
        variant
    }
    allele2_vars_plus_het_var = set(pair.allele2.defining_variants) | {
        variant
    }
    flattened = {
        allele
        for pair2 in bg.alleles[allele_state]
        for allele in pair2
        if pair2 != pair
    }
    hits = 0
    for a in flattened:
        if a not in pair:
            flat_vars = set(a.defining_variants)
            if (
                flat_vars == allele1_vars_plus_het_var
                or flat_vars == allele2_vars_plus_het_var
            ):
                hits += 1
    return hits


import operator
from functools import partial
from itertools import combinations, combinations_with_replacement
from typing import Iterable

from rbceq2.core_logic.alleles import Allele, BloodGroup, Pair
from rbceq2.core_logic.constants import LOW_WEIGHT, AlleleState
from rbceq2.core_logic.utils import (
    apply_to_dict_values,
    check_available_variants,
    chunk_geno_list_by_rank,
)
from icecream import ic


def sub_alleles(lst: tuple[Allele], allele_relationship: dict[str, bool]) -> bool:
    """look up a precomputed relationship with a special key - KN only

    Args:
        lst (list): A list of strings to be checked.

    Returns:
        list: A list of booleans indicating whether each element is in another element.
    """
    for allele1 in lst:
        for allele2 in lst:
            key = f"{allele1.genotype}_isin_{allele2.genotype}"
            if allele1 == allele2:
                continue
            # print(allele1, allele2)
            # if allele1.sub_type != allele2.sub_type:
            #     continue
            if allele_relationship[key]:
                return True
    return False


def all_hom_variants(all_homs: list[Allele], current_combo: tuple[Allele, ...]) -> bool:
    """
    Check if all homozygous alleles are present in the current combination.

    This function verifies whether all alleles designated as homozygous are
    present in the given combination of alleles.

    Args:
        all_homs (list[Allele]): A list of homozygous alleles to be checked. All_homs
        is a set of Alleles where all defining vars are HOM
        current_combo (tuple[Allele, ...]): A tuple of alleles representing the current
        combination.

    Returns:
        bool: True if the count DOES NOT match.

    Example:

    'possible': [
        Allele(genotype='KN*01.06',
                genotype_alt='.',
                defining_variants=frozenset({'1:207782856_A_G',
                                            '1:207782889_A_G',
                                            '1:207782916_A_T', hom
                                            '1:207782931_A_G'}), hom
                weight_geno=1000,
                reference=False,
        Allele(genotype='KN*01.07',
                genotype_alt='.',
                defining_variants=frozenset({'1:207782889_A_G',
                                            '1:207782916_A_T'}), hom
                weight_geno=1000,
                reference=False,
        Allele(genotype='KN*01.10',
                genotype_alt='.',
                defining_variants=frozenset({'1:207782916_A_T', hom
                                            '1:207782931_A_G'}), hom
                weight_geno=1000,
                reference=False,

    sample='140',
    variant_pool={'1:207782856_A_G': 'Heterozygous',
                '1:207782889_A_G': 'Heterozygous',
                '1:207782916_A_T': 'Homozygous',
                '1:207782931_A_G': 'Homozygous'},
    genotypes=['KN*01.06/KN*01.10'],
    phenotypes=[],
    """
    hom_count = 0
    current_sub_type = {allele.sub_type for allele in current_combo}
    assert len(current_sub_type) == 1
    current_sub_type_str = current_sub_type.pop()
    homs_of_same_sub_type = {
        allele for allele in all_homs if allele.sub_type == current_sub_type_str
    }
    if not homs_of_same_sub_type:
        return False
    for hom_allele in homs_of_same_sub_type:
        for allele in current_combo:
            if hom_allele == allele:
                hom_count += 1
            if hom_allele in allele:
                hom_count += 1
   
    return hom_count != len(all_homs)  # not less OR MORE


def mushed_vars(mushed_combo: Iterable[Allele]) -> set[str]:
    """Combine the defining_variants of all alleles in mushed_combo into a
    single set.

    Args:
        mushed_combo (Iterable[Allele]): An iterable of Allele objects whose
            defining_variants will be combined.

    Returns:
        set[str]: A set containing all unique defining_variants from the given
            Allele objects.
    """
    return set().union(
        *[mushed_allele.defining_variants for mushed_allele in mushed_combo]
    )


def can_co_exist(mushed_combo1: set[str], tmp_pool23: dict[str, int]) -> bool:
    """Check if the defining variants in 'mushed_combo1' can be satisfied by the
    'tmp_pool23' pool of available variants (count>0).

    Args:
        mushed_combo1 (set[str]): A set of defining variants.
        tmp_pool23 (dict[str, int]): A dictionary mapping variant names to their
            available counts.

    Returns:
        bool: True if all variants in mushed_combo1 are available in tmp_pool23
            with count > 0, False otherwise.
    """
    available_variants: set[str] = {
        variant for variant, count in tmp_pool23.items() if count > 0
    }
    intersection = mushed_combo1.intersection(available_variants)

    return len(intersection) == len(mushed_combo1)


def decide_if_co_existing(
    tmp_pool2: dict[str, int],
    combo1: tuple[Allele, ...],
    combo2: tuple[Allele, ...],
    mushed_combo1: set[str],
    co_existing: list[tuple[tuple[Allele, ...], tuple[Allele, ...]]],
) -> list[tuple[tuple[Allele, ...], tuple[Allele, ...]]]:
    """
    Decide if combo1 and combo2 can co-exist based on the variant pool and
    already discovered co-existing combos.

    Args:
        variant_pool: A dictionary of variant -> numeric count.
        combo1: A tuple of Alleles forming one combination.
        combo2: A tuple of Alleles forming the second combination.
        combined_variants: A set of defining variants from combo1 (used for can_co_exist).
        existing_pairs: The current list of co-existing combos to which we may add.

    Returns:
        The possibly updated list of existing_pairs.
    """
    check_vars_other_strand = partial(
        check_available_variants, 0, tmp_pool2, operator.gt
    )
    if all(
        all(check_vars_other_strand(allele2)) for allele2 in combo1
    ) and can_co_exist(mushed_combo1, tmp_pool2):
        # ie can they exist given other strand
        # if can_co_exist(mushed_combo1, tmp_pool2):
        combo_pair = (combo1, combo2)
        if combo_pair not in co_existing:
            co_existing.append(combo_pair)  # TODO - order for testing?

        # TODO sort on order in ISBT
    return co_existing


@apply_to_dict_values
def homs(
    bg: BloodGroup,
) -> BloodGroup:
    """Check all combinations of alleles against all other combinations to see if
    they can co-exist once the variant pool is reduced by the respective
    defining variants.

    Args:
        bg (BloodGroup): A BloodGroup object containing alleles, type, variant pool,
            and misc information.

    Returns:
        BloodGroup: The updated BloodGroup with modified alleles and misc values.
    """
    if bg.type != "KN":
        bg.alleles[AlleleState.CO] = None
        return bg
    unique_alleles = set(bg.alleles[AlleleState.FILT])

    check_hom_vars = partial(
        check_available_variants, 2, bg.variant_pool_numeric, operator.eq
    )
    bg.misc = {}  # TODO misc is a bit promiscuous
    bg.misc["homs"] = {  # URGENT TODO - this needs to be split by subtype!!
        allele
        for allele in unique_alleles
        if all(check_hom_vars(allele))
        if not allele.reference
    }

    return bg


@apply_to_dict_values
def max_rank(
    bg: BloodGroup,
) -> BloodGroup:
    """Check all combinations of alleles against all other combinations to see if
    they can co-exist once the variant pool is reduced by the respective defining
    variants.

    Args:
        bg (BloodGroup): A BloodGroup object containing alleles, type, variant pool,
            and misc information.

    Returns:
        BloodGroup: The updated BloodGroup with modified alleles and misc values.
    """
    if bg.type != "KN":
        bg.alleles[AlleleState.CO] = None
        return bg
    unique_alleles = set(bg.alleles[AlleleState.FILT])

    trumpiest = [
        allele.weight_geno
        for allele in list(bg.misc["homs"])
        + [a for a in unique_alleles if a.number_of_defining_variants == 1]
    ]

    bg.misc["max_rank"] = min(trumpiest) if trumpiest else LOW_WEIGHT

    return bg


@apply_to_dict_values
def prep_co_putative_combos(
    bg: BloodGroup,
    allele_relationships: dict[str, dict[str, bool]],
) -> BloodGroup:
    """Check all combinations of alleles against all other combinations
    to see if can co-exist , once the variant pool is reduced
    by the respective defining variants"""

    def make_allele_combos(
        flattened_alleles: list[Allele],
        homs: list[Allele],
        allele_relationship: dict[str, bool],
    ) -> list[tuple[Allele, ...]]:
        """Generate combinations of alleles, filtering out those that do not meet
        criteria.

        This function generates combinations of alleles from the provided list,
        filtering out combinations based on the presence of homozygous alleles,
        sub-alleles, and reference alleles.

        Args:
            flattened_alleles (List[Allele]): List of all alleles to consider.
            homs (List[Allele]): List of homozygous alleles.
            allele_relationship (Dict[str, bool]): Dictionary of allele relationships.
            bg_type (str): The blood group type.

        Returns:
            Set[Tuple[Allele, ...]]: Set of allele combinations that meet the criteria.
        """
        combos = []
        for i in range(1, len(flattened_alleles) + 1)[::-1]:
            for combo1 in combinations(flattened_alleles, i):
                if len({a.sub_type for a in combo1}) > 1:
                    continue
                #ic(333333, combo1, homs)
                if all_hom_variants(homs, combo1):
                    continue
                #ic(444444, combo1)
                if sub_alleles(combo1, allele_relationship):
                    continue
                #ic(555555, combo1)
                if any(a.reference for a in combo1):
                    continue
                #ic(666666, combo1)
                ranked_combo1 = tuple(chunk_geno_list_by_rank(combo1)[0])

                assert ranked_combo1 not in combos
                combos.append(ranked_combo1)
        #ic(8888888, combos)
        return combos

    if bg.type != "KN":
        bg.alleles[AlleleState.CO] = None
        return bg
    
    unique_alleles = sorted(
        set(bg.alleles[AlleleState.FILT]), key=lambda allele: allele.genotype
    )
    bg.misc["combos"] = make_allele_combos(
        unique_alleles, bg.misc["homs"], allele_relationships[bg.type]
    )
    #ic(555555666666665,set(bg.alleles[AlleleState.FILT]))
    return bg


@apply_to_dict_values
def add_co_existing_alleles(
    bg: BloodGroup,
) -> BloodGroup:
    """Check all combinations of alleles against all other combinations
    to see if can co-exist , once the variant pool is reduced
    by the respective defining variants

    Identify and store combinations of alleles that can co-exist in a BloodGroup.

    This function is specifically designed for blood groups of type "KN". It examines
    all pairwise combinations (including self-combinations) in `bg.misc["combos"]`.
    For each pair, the function reduces the available variant pool by the defining
    variants of the second combination, then decides whether both sets of alleles
    can co-exist without conflict. If so, the combination is added to the
    `AlleleState.CO` list.

    Args:
        bg (BloodGroup):
            The BloodGroup object whose alleles are being checked for co-existence.

    Returns:
        BloodGroup:
            The input BloodGroup object with its `AlleleState.CO` attribute updated
            to either a list of co-existing allele combinations (if any) or `None`
            (if no viable combinations are found).

    Raises:
        KeyError:
            If an expected variant is missing from `bg.variant_pool_numeric`.
    """

    if bg.type != "KN":
        bg.alleles[AlleleState.CO] = None
        return bg
    co_existing = []
    for combo1, combo2 in combinations_with_replacement(bg.misc["combos"], 2):
        # This includes pairs like (A, A), (A, B), (B, B), but not both (A, B) and (B, A)
        # ic(combo1, combo2)
        mushed_combo1 = mushed_vars(combo1)
        tmp_pool2 = bg.variant_pool_numeric
        for variant_on_other_strand in mushed_vars(combo2):
            tmp_pool2[variant_on_other_strand] -= 1
        co_existing = decide_if_co_existing(
            tmp_pool2, combo1, combo2, mushed_combo1, co_existing
        )
        # TODO else - record why combo failed in bg
    if co_existing:
        bg.alleles[AlleleState.CO] = co_existing
    else:
        bg.alleles[AlleleState.CO] = None
    return bg


@apply_to_dict_values
def add_co_existing_allele_and_ref(
    bg: BloodGroup,
    reference_alleles: dict[str, Allele],
) -> BloodGroup:
    """Check all combinations of alleles against all other combinations
    to see if can co-exist , once the variant pool is reduced
    by the respective defining variants

    Evaluate allele combinations against a reference allele for co-existence.

    This function also targets blood groups of type "KN". For each combination
    in `bg.misc["combos"]`, it checks whether the allele set (combo1) can coexist
    with the specified reference allele (drawn from `reference_alleles`). It does so
    by reducing the `variant_pool_numeric` according to the reference allele's defining
    variants, then verifying if both sets remain viable.

    Args:
        bg (BloodGroup):
            The BloodGroup object whose alleles are being tested for co-existence
            with a reference allele.
        reference_alleles (dict[str, Allele]):
            A dictionary mapping blood group types to their corresponding reference
            Allele object.

    Returns:
        BloodGroup:
            The updated BloodGroup object. The `AlleleState.CO` attribute is assigned
            a list of (combination, reference) pairs that can coexist, or `None` if
            no valid pairs are found.

    Raises:
        KeyError:
            If an expected variant is missing from `bg.variant_pool_numeric` or
            `reference_alleles`.
    """
    if bg.type != "KN":
        bg.alleles[AlleleState.CO] = None
        return bg

    if bg.alleles[AlleleState.CO] is not None:
        co_existing = bg.alleles[AlleleState.CO]
    else:
        co_existing = []
    for combo1 in bg.misc["combos"]:
        mushed_combo1 = mushed_vars(combo1)
        if bg.misc["homs"]:
            continue
        if bg.misc["max_rank"] < min([allele.weight_geno for allele in combo1]):
            continue
        tmp_pool2 = bg.variant_pool_numeric
        ref = reference_alleles[bg.type]
        # some refs have defining vars (ie ABO*A1.01 has
        # 261delG/136132908_T_TC)
        for variant_on_other_strand in ref.defining_variants:
            if variant_on_other_strand in tmp_pool2:
                tmp_pool2[variant_on_other_strand] -= 1
        co_existing = decide_if_co_existing(
            tmp_pool2, combo1, (ref,), mushed_combo1, co_existing
        )
    if co_existing:
        bg.alleles[AlleleState.CO] = co_existing
    else:
        bg.alleles[AlleleState.CO] = None
    return bg


@apply_to_dict_values
def filter_redundant_pairs(bg: BloodGroup) -> BloodGroup:
    """
    Filter out duplicate pairs, considering both original and reversed forms.

    Args:
        co_existing (list[str]): List of string pairs separated by '/'.

    Returns:
        List of unique combo pairs.
    """

    if bg.alleles[AlleleState.CO] is None:
        return bg
    seen = set()
    result = []
    for combo1, combo2 in bg.alleles[AlleleState.CO]:
        geno1, geno2 = geno_str(combo1), geno_str(combo2)
        fwd = f"{geno1}/{geno2}"
        rev = f"{geno2}/{geno1}"
        if fwd not in seen and rev not in seen:
            seen.update({fwd, rev})
            result.append((combo1, combo2))
    bg.alleles[AlleleState.CO] = result

    return bg


@apply_to_dict_values
def mush(bg: BloodGroup) -> BloodGroup:
    """
    Mushes co-existing alleles into a new allele:
      - Iterates over each 'combo' in bg.alleles[AlleleState.CO]
      - For each sub-combo in combos:
          if len(combo) > 1 -> we unify numeric + alphanumeric phenotypes
                               and produce a new "mushed" Allele
          if len(combo) == 1 -> we keep the single allele as-is
      - Then each mushed_pair must have exactly two Alleles to form a Pair.
      - Raises ValueError if there's a conflict or if the pair doesn't have exactly two.

    Args:
        bg: A BloodGroup object whose 'CO' (co-existing) combos we want to merge.

    Returns:
        The same BloodGroup, with bg.alleles[AlleleState.CO] replaced by
        new Pairs of Alleles or set to None if none remain.

    Raises:
        ValueError: if any conflict arises or a pair isn't exactly two Alleles.
    """

    def _check_conflict(
        expressions: dict[str, str], name: str, expr: str
    ) -> dict[str, str]:
        """Raise ValueError if 'name' is already in 'expressions' with a different expr."""
        if name in expressions and expressions[name] != expr:
            raise ValueError(
                f"Conflicting expressions for antigen '{name}': "
                f"'{expressions[name]}' and '{expr}'"
            )
        expressions[name] = expr
        return expressions

    result = []
    if bg.alleles[AlleleState.CO] is None:
        return bg
    for combos in bg.alleles[AlleleState.CO]:
        mushed_pair = []
        for combo in combos:
            if len(combo) > 1:
                # Process multiple alleles in a combo
                geno_numeric = geno_str(combo)

                # Process pheno_numeric
                phenotypes = [allele.phenotype.split(":")[1] for allele in combo]
                antigens = []
                for phenotype in phenotypes:
                    antigens.extend(phenotype.split(","))

                antigen_expressions: dict[str, str] = {}
                for antigen in antigens:  # TODO should I use Ant objects here!!!!????
                    if antigen.startswith("-"):
                        expression, name = "-", antigen[1:]
                    else:
                        expression, name = "+", antigen
                    antigen_expressions = _check_conflict(
                        antigen_expressions, name, expression
                    )

                unique_antigens = []
                for name, expression in antigen_expressions.items():
                    if expression == "-":
                        unique_antigens.append(f"-{name}")
                    else:
                        unique_antigens.append(name)
                unique_antigens.sort(key=lambda x: x.lstrip("-"))

                pheno_numeric = f"{bg.type}:{','.join(unique_antigens)}"
                if "1w," in pheno_numeric and "1," in pheno_numeric:
                    pheno_numeric = pheno_numeric.replace("1w,", "")  # Helgeson

                # Process pheno_alphanumeric
                phenotypes_alt = [allele.phenotype_alt for allele in combo]
                antigens_alt = []
                for phenotype_alt in phenotypes_alt:
                    antigens_alt.extend(phenotype_alt.split(","))

                antigen_expressions_alt: dict[str, str] = {}
                for antigen in antigens_alt:
                    if "-" in antigen:
                        expression = "-"
                        name = antigen.replace("-", "")
                    else:
                        expression = "+"
                        name = antigen.replace("+", "")
                    antigen_expressions_alt = _check_conflict(
                        antigen_expressions_alt, name, expression
                    )

                unique_antigens_alt = []
                for name, expression in antigen_expressions_alt.items():
                    if expression == "-":
                        if ")" in name:
                            unique_antigens_alt.append(name.replace(")", "-)"))
                        else:
                            unique_antigens_alt.append(f"{name}-")
                    elif ")" in name:
                        unique_antigens_alt.append(name.replace(")", "+)"))
                    else:
                        unique_antigens_alt.append(f"{name}+")
                unique_antigens_alt.sort(key=lambda x: x.rstrip("+-"))

                pheno_alphanumeric = ",".join(unique_antigens_alt)
                if "Kn(aw+)," in pheno_alphanumeric and "Kn(a+)," in pheno_alphanumeric:
                    pheno_alphanumeric = pheno_alphanumeric.replace(
                        "Kn(aw+),", ""
                    )  # Helgeson
                mushed_pair.append(
                    make_mushed_allele(
                        combo, geno_numeric, pheno_numeric, pheno_alphanumeric
                    )
                )
            elif len(combo) == 1:
                # Single allele, append as is
                mushed_pair.append(combo[0])

        # Ensure mushed_pair has exactly two alleles
        if len(mushed_pair) == 2:
            result.append(Pair(*mushed_pair))
        else:
            # Handle cases where there's only one allele
            # You can decide how to handle this based on your requirements
            # For now, let's raise an error
            raise ValueError("Each mushed_pair must contain exactly two alleles.")
    if result:
        bg.alleles[AlleleState.CO] = result
    else:
        bg.alleles[AlleleState.CO] = None

    return bg


def geno_str(combo: tuple[Allele, ...]) -> str:
    """Build the genotype string by sorting on allele.genotype and joining with '+'.

    Args:
        combo (tuple[Allele, ...]): A tuple of Allele objects.

    Returns:
        str: A string with sorted allele.genotype values concatenated by '+'.
    """
    return "+".join(sorted([a.genotype for a in combo]))


def make_mushed_allele(
    combo: tuple[Allele, ...], geno: str, pheno_numeric: str, pheno_alpha: str
) -> Allele:
    """Mush co existing alleles into one new allele.

    Args:
        combo (tuple[Allele, ...]): A tuple of Allele objects to be merged.
        geno (str): The genotype string for the new allele.
        pheno_numeric (str): The numeric phenotype string for the new allele.
        pheno_alpha (str): The alphanumeric phenotype string for the new allele.

    Returns:
        Allele: A new Allele object with merged attributes from the input alleles.
    """
    for pheno in [pheno_numeric, pheno_alpha]:
        assert ",," not in pheno
    return Allele(
        genotype=geno,
        phenotype=pheno_numeric,
        genotype_alt="mushed",
        phenotype_alt=pheno_alpha,
        defining_variants=frozenset.union(*[a.defining_variants for a in combo]),
        null=False,
        weight_geno=min([a.weight_geno for a in combo]),
        sub_type="/".join({a.sub_type for a in combo}),
    )


@apply_to_dict_values
def list_excluded_co_existing_pairs(
    bg: BloodGroup,
    reference_alleles: dict[str, Allele],
) -> BloodGroup:
    """Check all combinations of alleles against all other combinations to see if
    they can co-exist once the variant pool is reduced by the respective defining
    variants.

    Args:
        bg (BloodGroup): A BloodGroup object containing alleles, type, and misc
            information.
        reference_alleles (dict[str, Allele]): A dictionary mapping blood group types
            to their reference Allele.

    Returns:
        BloodGroup: The updated BloodGroup with filtered out co-existing pairs.
    """
    if bg.type != "KN":
        bg.alleles[AlleleState.CO] = None
        return bg
    tested = []
    ref = reference_alleles[bg.type]
    for combo1 in bg.misc["combos"]:
        for combo2 in bg.misc["combos"]:
            tested.append(
                Pair(
                    make_mushed_allele(combo1, geno_str(combo1), "numeric", "alpha"),
                    make_mushed_allele(combo2, geno_str(combo2), "numeric", "alpha"),
                )
            )
        tested.append(
            Pair(ref, make_mushed_allele(combo1, geno_str(combo1), "numeric", "alpha"))
        )

    bg.filtered_out[AlleleState.CO] = [
        pair for pair in tested if pair not in bg.alleles[AlleleState.CO]
    ]
    # print(1111112,bg.alleles[AlleleState.CO], '\n', bg.filtered_out[AlleleState.CO])

    return bg



from __future__ import annotations

import copy
from collections import Counter, defaultdict
from functools import partial
from itertools import product
from typing import TYPE_CHECKING
from venv import logger

from rbceq2.core_logic.constants import ANTITHETICAL, AlleleState, BgName, PhenoType
from rbceq2.core_logic.data_procesing import apply_to_dict_values
from rbceq2.core_logic.utils import (
    BeyondLogicError,
)
from icecream import ic

from rbceq2.core_logic.alleles import Allele

from . import antigens as an

if TYPE_CHECKING:
    import pandas as pd

    from rbceq2.core_logic.alleles import BloodGroup, Pair


def fut_helper(res: dict[str, BloodGroup]):
    """Generate all possible pairs of alphanumeric phenotypes from FUT1 and FUT2.

    This helper function extracts the alphanumeric phenotype values from the FUT1 and
    FUT2 BloodGroup objects, then returns the Cartesian product of these phenotype sets
    as a list of tuples.

    Args:
        res (dict[str, BloodGroup]): Dictionary of blood group results containing at
            least keys "FUT1" and "FUT2". Each corresponding value is a BloodGroup object
            with a 'phenotypes' attribute, where phenotypes are categorized by PhenoType.

    Returns:
        list[tuple[str, str]]: A list of tuples, where each tuple represents a pair of
            phenotype values, one from FUT1 and one from FUT2.
    """
    fut1 = res["FUT1"]
    fut2 = res["FUT2"]
    pheno_key: PhenoType = PhenoType.alphanumeric
    fut1s = set(fut1.phenotypes[pheno_key].values())
    fut2s = set(fut2.phenotypes[pheno_key].values())
    fut1_and_2 = list(product(fut1s, fut2s))
    return fut1_and_2


def FUT1(res: dict[str, BloodGroup]) -> dict[str, BloodGroup]:
    """Update FUT1 phenotypes based on interactions between FUT1 and FUT2.

    In the context of fucosyltransferase blood group antigens, FUT1 (H), FUT2 (Se),
    and FUT3 (Le) have interrelated phenotypes. This function updates the alphanumeric
    phenotypes of the FUT1 BloodGroup by combining information from FUT1 and FUT2.
    For each phenotype in FUT1, it finds all phenotype pairs from FUT1 and FUT2 that
    include the original FUT1 phenotype, concatenates them, and then stores the result
    as the new phenotype value.

    Args:
        res (dict[str, BloodGroup]): Dictionary of blood group results, where keys
            are 'FUT1', 'FUT2', and 'FUT3', each mapped to their respective BloodGroup
            data.

    Returns:
        dict[str, BloodGroup]: The updated dictionary with modified FUT1 phenotypes.
    """

    new_phenos = {}
    fut1_and_2 = fut_helper(res)
    for pair, pheno in res["FUT1"].phenotypes[PhenoType.alphanumeric].items():
        new = []
        for combo in fut1_and_2:
            if pheno in combo:
                new.append(",".join(combo))
        new_phenos[pair] = "/".join(sorted(new))

    res["FUT1"].phenotypes[PhenoType.alphanumeric] = new_phenos

    return res


def FUT3(res: dict[str, BloodGroup]) -> dict[str, BloodGroup]:
    """
    H = FUT1
    Se = FUT2
    Le = FUT3

    In individuals with an active FUT2 (Secretor or SE) gene, which encodes a fully
    active α(1,2)-fucosyltransferase (see H Blood Group System), predominantly
    Leb (and related Lewis antigens depending on ABO group, e.g. ALeb in group
    A) is made alongside trace amounts of Lea. The trace amounts of Lea produced
    are typically undetectable using serological methods, and usually, a Le(a–b+)
    phenotype is reported.

    functional == active???!!!

    LE	When FUT2 is H+ Se+ and FUT3 is functional = Le(a-b+)
        When FUT2 is H+w Se+ and FUT3 is functional =Le(a+b+)
        When FUT2 is homozygous null (Se-) and FUT3 is functional = Le(a+b-)
        When FUT3 is homozygous null, regardless of FUT2 functionality = Le(a-b-)
    H	FUT1: RBC expresssion
        FUT2: Secretor status - Report phenotype as Se+/-

    Args:
        res (dict): Dictionary of blood group results, where keys are 'FUT1', 'FUT2',
        and 'FUT3', each with their respective `BloodGroup` data.

    Returns:
        dict: Updated dictionary with modified FUT3 phenotypes based on FUT1 and FUT2
        interactions.
    """

    fut1_and_2 = fut_helper(res)

    new_phenos = {}
    for pair, pheno in res["FUT3"].phenotypes[PhenoType.alphanumeric].items():
        new_pheno = []
        if "active" in pheno.lower():
            for combo in fut1_and_2:
                if "Se-" in combo:
                    new_pheno.append("Le(a+b-)")
                elif "H+w" in combo and "Se+" in combo:
                    new_pheno.append("Le(a+b+)")
                elif "H+" in combo and any(se in combo for se in ["Se+", "Se+w"]):
                    new_pheno.append("Le(a-b+)")
                else:
                    raise BeyondLogicError(
                        message="Unexpected FUT3 value.",
                        context=f"Received value: {combo}",
                    )
        elif "Le(a-b-)" in pheno:
            new_pheno.append("Le(a-b-)")
        else:
            raise BeyondLogicError(
                message="Unexpected FUT3 pheno.", context=f"Received value: {pheno}"
            )

        new_phenos[pair] = "/".join(sorted(set(new_pheno)))

    res["FUT3"].phenotypes[PhenoType.alphanumeric] = new_phenos

    return res


def choose_class_type(bg_type, ant_type):
    """Choose the antigen class based on blood group and phenotype type.

    Args:
        bg_type (BgName): The blood group type.
        ant_type (PhenoType): The phenotype type, either alphanumeric or numeric.

    Returns:
        type: The antigen class corresponding to the given types.
    """
    class_types = {
        PhenoType.alphanumeric: {
            BgName.XG: an.AlphaNumericAntigenXG,
            BgName.VEL: an.AlphaNumericAntigenVel,
            BgName.GYPA: an.AlphaNumericAntigenMNS,
            BgName.GYPB: an.AlphaNumericAntigenMNS,
            BgName.ABO: an.AlphaNumericAntigenABO,
            BgName.RHCE: an.AlphaNumericAntigenRHCE,
            BgName.RHD: an.AlphaNumericAntigenRHD,
        },
        PhenoType.numeric: {
            BgName.RHCE: an.NumericAntigenRHCE,
            BgName.VEL: an.NumericAntigenVel,
        },
    }
    return (
        class_types[ant_type].get(bg_type, an.NumericAntigen)
        if ant_type == PhenoType.numeric
        else class_types[ant_type].get(bg_type, an.AlphaNumericAntigen)
    )


def make_values_dict(
    values_strs: list[str], ant_type: PhenoType, bg_type: BgName
) -> dict[str, list[an.Antigen]] | None:
    """Construct a dictionary of Antigen objects from antigen value strings.

    This function processes a list of antigen strings and creates a dictionary where the keys
    are antigen base names and the values are lists of corresponding Antigen objects. The antigen
    objects are instantiated using a class determined by the phenotype type (ant_type) and blood
    group type (bg_type).

    Processing steps:
      1. If any string in values_strs contains a period ('.'), return None. This is because some
         blood groups may not have consistent numeric or alphanumeric allele representations.
      2. Determine the antigen class to use by calling `choose_class_type(bg_type, ant_type)`.
      3. Modify the input strings:
           - For numeric antigens, split each string by ":" and use the second part.
           - For alphanumeric antigens, use the string as is.
      4. Concatenate the modified strings, split by commas, and count the occurrences using a Counter.
      5. For each unique antigen string and its count:
           - Create an antigen object with:
               - given_name set to the antigen string.
               - expressed flag determined by the absence of '-' in the antigen string.
               - homozygous flag set to True if the count equals 2.
               - antithetical_relationships set based on the global ANTITHETICAL dictionary for the given
                 phenotype and blood group types.
           - Assert that the antigen's base_name is not empty.
           - Append the antigen to the dictionary under its base_name.

    Args:
        values_strs (list[str]): A list of strings representing antigen values.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) of the antigen.
        bg_type (BgName): The blood group type used to determine the antigen class and relationships.

    Returns:
        dict[str, list[an.Antigen]] | None: A dictionary mapping antigen base names to lists of Antigen objects,
        or None if the input format is invalid.
    """
    antigens = defaultdict(list)
    # all dicts are ordered
    if "." in values_strs:
        return None  # some BGs just don't have numeric for all or alpha for all alleles
    ant_class = choose_class_type(bg_type, ant_type)
    values_strs_mod = []
    for values_str in values_strs:
        values_strs_mod.append(
            values_str.split(":")[1] if ant_type == PhenoType.numeric else values_str
        )

    # NB order only matters for ref, where there's just one str (Counter is ordered)
    for ant_str, count in Counter(",".join(values_strs_mod).split(",")).items():
        ant = ant_class(
            given_name=ant_str,
            expressed="-" not in ant_str,
            homozygous=count == 2,
            antithetical_relationships=ANTITHETICAL[ant_type].get(bg_type, {}),
        )
        assert ant.base_name != ""
        antigens[ant.base_name].append(ant)

    return antigens


@apply_to_dict_values
def add_ref_phenos(bg: BloodGroup, df: pd.DataFrame) -> BloodGroup:
    """Add reference phenotypes to the BloodGroup from a DataFrame.

    This function filters the DataFrame for the reference genotype corresponding
    to the BloodGroup type, then constructs alphanumeric and numeric phenotype
    dictionaries using make_values_dict. These are stored in the bg.misc attribute.

    Args:
        bg (BloodGroup): The BloodGroup object to update.
        df (pd.DataFrame): A DataFrame containing phenotype data. Must include
            'type', 'Reference_genotype', 'Phenotype_alt', and 'Phenotype' columns.

    Returns:
        BloodGroup: The updated BloodGroup with reference phenotypes added.

    Raises:
        AssertionError: If the filtered DataFrame does not have exactly one row.
    """
    df["type"] = df["type"].astype("category")
    df_ref = df.loc[(df["Reference_genotype"] == "Yes") & (df["type"] == bg.type)]

    try:
        assert df_ref.shape == (1, 20)
    except AssertionError:
        ic(df_ref.shape, df_ref)
    bg.misc = {}
    bg.misc["ref_PhenoType.alphanumeric"] = make_values_dict(
        [df_ref.iloc[0]["Phenotype_alt"]],
        PhenoType.alphanumeric,
        BgName.from_string(bg.type),
    )
    bg.misc["ref_PhenoType.numeric"] = make_values_dict(
        [df_ref.iloc[0]["Phenotype"]], PhenoType.numeric, BgName.from_string(bg.type)
    )

    return bg


@apply_to_dict_values
def instantiate_antigens(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Instantiate antigen objects for a BloodGroup based on allele phenotype changes.

    This function iterates over allele pairs in the BloodGroup (using either the CO or
    NORMAL state, as determined by the inner helper) and generates a dictionary of antigens.
    For each allele pair, phenotype changes are extracted based on the phenotype type (numeric
    or alphanumeric). If the phenotype changes indicate that the pair is informative (i.e. not
    just missing data represented by '.'), the function uses a pre-filled partial of
    make_values_dict to generate antigen objects. The resulting dictionary is stored in the
    bg.misc attribute under a key formatted as "antigens_{ant_type.name}".

    Args:
        bg (BloodGroup): A BloodGroup object containing allele pairs and associated data.
        ant_type (PhenoType): The phenotype type, which influences the phenotype attribute
            to be used and the antigen class to instantiate.

    Returns:
        BloodGroup: The updated BloodGroup with instantiated antigens added to bg.misc.
    """

    def select_allele_state(bg):
        """Select the allele state to use for antigen instantiation.

        For blood groups of type "KN", if co-existing alleles (CO) are present, then
        AlleleState.CO is returned. Otherwise, AlleleState.NORMAL is used.

        Args:
            bg (BloodGroup): The BloodGroup object.

        Returns:
            AlleleState: The allele state (CO or NORMAL) to be used.
        """
        if AlleleState.CO in bg.alleles and bg.alleles[AlleleState.CO] is not None:
            assert bg.type == "KN"
            return AlleleState.CO
        return AlleleState.NORMAL

    def get_pheno_changes(current_pair: Pair, ant_type: PhenoType) -> list[str]:
        """Extract phenotype changes from an allele pair based on phenotype type.

        For numeric phenotypes, the 'phenotype' attribute is used; for alphanumeric,
        'phenotype_alt' is used. When the BloodGroup type is "ABO", the function uses
        alleles with expressed phenotypes; otherwise, it uses all alleles in the pair.

        Args:
            current_pair (Pair): The allele pair from which to extract phenotype changes.
            ant_type (PhenoType): The phenotype type (numeric or alphanumeric).

        Returns:
            list[str]: A list of phenotype strings derived from the allele pair.
        """
        def filter_abo_types(alleles: list[Allele]) -> frozenset[str]:
            """
            Filters ABO types based on the presence of 'ABO*O.'.

            Args:
                Allele1: The first ABO type string.
                Allele2: The second ABO type string.

            Returns:
                A frozenset containing:
                - Both strs if both contain 'ABO*O.'.
                - Only the str that does NOT contain 'ABO*O.' if one of them does.
                - Both strs if neither contains 'ABO*O.'.
            """
            if len(alleles) == 1:
                return alleles
            elif len(alleles) == 2:
                a1, a2 = alleles
                contains_o1 = "ABO*O." in a1.genotype
                contains_o2 = "ABO*O." in a2.genotype

                if contains_o1 and contains_o2:
                    return frozenset([a1, a2])
                elif contains_o1 and not contains_o2:
                    return frozenset([a2])
                elif not contains_o1 and contains_o2:
                    return frozenset([a1])
                else:  # Neither contains 'ABO*O.'
                    return frozenset([a1, a2])
            else:
                raise ValueError('ABO allele count wrong')
    
        phenotype_attr = (
            "phenotype" if ant_type == PhenoType.numeric else "phenotype_alt"
        )
        alleles_to_use = (
            filter_abo_types(current_pair.alleles) 
            if bg.type == "ABO"
            else current_pair.alleles
        )

        return [getattr(allele, phenotype_attr) for allele in alleles_to_use]

    pair_antigens = {}
    make_values_dict_pre_filled = partial(
        make_values_dict, ant_type=ant_type, bg_type=BgName.from_string(bg.type)
    )
    for pair in bg.alleles[select_allele_state(bg)]:
        pheno_changes = get_pheno_changes(pair, ant_type)
        if set(pheno_changes) != {"."}:  # 1 or more defininitions missing
            if pair.allele1 == pair.allele2:
                pheno_changes *= 2
            pair_antigens[pair] = make_values_dict_pre_filled(values_strs=pheno_changes)
            # TODO need to do this specifically for the PhenoType and only if the other one isn't {"."}
            # TODO replace bg.type globally

    bg.misc[f"antigens_{ant_type.name}"] = pair_antigens
    
    return bg


@apply_to_dict_values
def get_phenotypes1(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Integrate antigen information with reference phenotypes for a BloodGroup.

    This function retrieves the reference antigen data from bg.misc using the key
    corresponding to the provided phenotype type. It then iterates over the antigen
    pairs stored in bg.misc for that phenotype type and merges each pair with the
    reference data when necessary. The merging behavior depends on whether the allele
    is homozygous, heterozygous, or has antithetical antigens. The combined antigen
    information is stored back in bg.misc under a key formed by concatenating
    "antigens_and_ref_" with the antigen type name.

    Args:
        bg (BloodGroup): The BloodGroup object containing antigen and reference data.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) for which
            antigen data is being processed.

    Returns:
        BloodGroup: The updated BloodGroup with combined antigen and reference
            information in its misc attribute.

    Raises:
        BeyondLogicError: If an unexpected configuration of allele antigens is encountered.
    """
    reference = bg.misc[f"ref_{str(ant_type)}"]
    if reference is None:
        return bg
    else:
        ref = copy.deepcopy(reference)

    d = {}
    for pair, antigens in bg.misc[f"antigens_{ant_type.name}"].items():
        if antigens is None:
            continue
        antigens_with_ref_if_needed = {}
        for ant_pos, allele_antigens in antigens.items():
            if (
                bg.type in ["ABO"]
                or allele_antigens[0].homozygous
                and len(allele_antigens) == 1
            ):
                antigens_with_ref_if_needed[ant_pos] = allele_antigens
            elif len(allele_antigens) == 2:
                assert all(not allele.homozygous for allele in allele_antigens), (
                    "Expected both alleles to be heterozygous"
                )
                antigens_with_ref_if_needed[ant_pos] = allele_antigens
            elif len(allele_antigens) == 1:
                ant = allele_antigens[0]
                assert not ant.homozygous
                add_ref = True
                if ant.antithetical_antigen is not None:
                    for anti_ant in ant.antithetical_antigen:
                        if anti_ant.base_name in antigens.keys():
                            add_ref = False
                if add_ref:
                    ant_pair = allele_antigens + ref[ant_pos]
                    antigens_with_ref_if_needed[ant_pos] = ant_pair
                else:
                    antigens_with_ref_if_needed[ant_pos] = allele_antigens
            else:
                raise BeyondLogicError(
                    message="Unexpected allele_antigens.",
                    context=f"Received value: {allele_antigens}",
                )
            d[pair] = antigens_with_ref_if_needed

    bg.misc[f"antigens_and_ref_{ant_type.name}"] = d

    return bg


@apply_to_dict_values
def get_phenotypes2(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Merge antigen phenotype options and update BloodGroup phenotype mapping.

    This function integrates antigen data with reference information by merging
    phenotype options for each allele pair. It iterates over the antigen pairs stored
    in bg.misc under the key "antigens_and_ref_{ant_type.name}" and applies merging rules
    based on antigen equality, given names, and weight comparisons. The resulting merged
    antigen data is then stored in bg.phenotypes for the provided phenotype type.

    Args:
        bg (BloodGroup): A BloodGroup object containing antigen, reference, and phenotype data.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) to process.

    Returns:
        BloodGroup: The updated BloodGroup with merged phenotype information.

    Raises:
        BeyondLogicError: If an unexpected antigen pattern is encountered during merging.
    """
    reference = bg.misc[f"ref_{str(ant_type)}"]
    if reference is None:
        return bg

    for pair, antigens_with_ref_if_needed in bg.misc[
        f"antigens_and_ref_{ant_type.name}"
    ].items():
        if antigens_with_ref_if_needed is None:
            continue

        merged_pheno2 = []
        for _, options in antigens_with_ref_if_needed.items():
            if len(options) == 2:
                ant1, ant2 = options
                assert ant1.base_name == ant2.base_name
                if ant1 == ant2 and ant1.given_name != ant2.given_name:
                    merged_pheno2.append(ant1)
                    merged_pheno2.append(ant2)
                elif ant1 == ant2 and ant1.given_name == ant2.given_name:
                    if ant1.expressed and ant2.expressed:
                        ant1.homozygous = (
                            True  # TODO fix - can freeze class withotut this
                        )
                    merged_pheno2.append(ant1)
                elif ant1 > ant2:
                    if ant1.expressed and ant2.expressed:
                        ant2.homozygous = True
                    merged_pheno2.append(ant2)
                elif ant2 > ant1:
                    if ant1.expressed and ant2.expressed:
                        ant1.homozygous = True
                    merged_pheno2.append(ant1)
                else:
                    raise BeyondLogicError(
                        message="Unexpected antigne pattern.",
                        context=f"Received value: {options}",
                    )
            elif len(options) == 1:
                merged_pheno2.append(options[0])
            else:
                raise BeyondLogicError(
                    message="Unexpected antigne pattern.",
                    context=f"Received value: {options}",
                )

        bg.phenotypes[ant_type][pair] = merged_pheno2
        
    return bg


def count_expressed_ants(current_ant, current_base_names):
    """Count the number of expressed antigens including antithetical ones.

    The count is determined by:
      - Adding 1 if the current antigen is expressed.
      - Multiplying by 2 if the current antigen is homozygous.
      - For each antithetical antigen in current_ant.antithetical_antigen,
        if it exists in current_base_names and is expressed, add 1 if heterozygous,
        or 2 if homozygous.

    Args:
        current_ant (an.Antigen): The antigen to evaluate.
        current_base_names (dict[str, an.Antigen]): A mapping from antigen base names to
            corresponding antigen objects.

    Returns:
        int: The total number of expressed antigens.
    """
    number_of_expressed_ants = 1 if current_ant.expressed else 0
    if current_ant.homozygous:
        number_of_expressed_ants *= 2

    for antithetical_ant in current_ant.antithetical_antigen:
        ant_to_count = current_base_names.get(antithetical_ant.base_name)
        if ant_to_count is None:
            continue
        if ant_to_count.expressed:
            if ant_to_count.homozygous:
                number_of_expressed_ants += 2
            else:
                number_of_expressed_ants += 1

    return number_of_expressed_ants


@apply_to_dict_values
def internal_anithetical_consistency_HET(
    bg: BloodGroup, ant_type: PhenoType
) -> BloodGroup:
    """
    Ensure correct antigen expression for heterozygous antithetical pairs.

    In certain cases, when an antithetical pair is present and only one is
    marked with '-', the code sets it to positive if heterozygous. This often
    arises in tricky alleles like KEL*02.03 (which flips several antigens).

    Steps:
        1. count expressed antithetical antigens (from antigens not
        ant.antithetical_antigen)
        2. If two are already expressed, do nothing.
        3. If not, retrieve reference states to correct expression for the
           missing antigens.

    *NB some antigens have 3 antitheticals described ie (3,4, and 21 are
    anithetical in KEL, this is well tested). Others seem to have 3 but only
    2 are known - listed in sudo nulls (less well tested)

    Args:
        bg (BloodGroup): The blood group data, containing phenotypes and references.
        ant_type (PhenoType): The phenotype group to process.

    Returns:
        A potentially updated BloodGroup with corrected antithetical expressions.

    """

    def check_expression(antigen):
        ref_ant = reference[antigen.base_name][0]
        if ref_ant.expressed:
            new_antigens.append(ref_ant)
            already_checked.add(ref_ant.base_name)

    new_phenos = []
    ref = bg.misc[f"ref_{str(ant_type)}"]
    if ref is not None:
        reference = copy.deepcopy(ref)
    else:
        return bg

    for pair, antigens in bg.phenotypes[ant_type].items():
        null = pair.allele1.null or pair.allele2.null
        already_checked = set()
        new_antigens = []
        if pair.allele1.phenotype == ".":
            new_phenos.append((pair, antigens))
            continue
        if (
            "N." in pair.allele1.genotype.upper()
            or "N." in pair.allele2.genotype.upper()
        ):
            # TODO think about how to handle nulls better
            new_phenos.append((pair, antigens))
            continue
        base_names = {ant2.base_name: ant2 for ant2 in antigens}

        for ant in antigens:
            if ant.base_name in already_checked:
                continue
            if ant.antithetical_antigen:
                no_expressed = count_expressed_ants(ant, base_names)
                # no change
                if no_expressed == 2:
                    for ant2 in antigens:
                        if ant2.base_name == ant.base_name:
                            new_antigens.append(ant2)
                            already_checked.add(ant2.base_name)
                    continue
                # add expressed ants
                for ant2 in antigens:
                    if ant2.base_name == ant.base_name and ant2.expressed:
                        new_antigens.append(ant2)
                        already_checked.add(ant2.base_name)
                # add expressed ref ants
                if ant.base_name not in already_checked:
                    check_expression(ant)
                for antithetical_ant in ant.antithetical_antigen:
                    if antithetical_ant.base_name in already_checked:
                        continue
                    check_expression(antithetical_ant)
            else:
                new_antigens.append(ant)
            already_checked.add(ant.base_name)

        base_names_new = {ant2.base_name: ant2 for ant2 in new_antigens}
        for ant_base_name, ant in base_names.items():
            if ant_base_name not in base_names_new:
                assert not ant.expressed
                new_antigens.append(ant)
        new_phenos.append((pair, new_antigens))
        base_names_new = {ant2.base_name: ant2 for ant2 in new_antigens}

        for ant in base_names:
            assert ant in base_names_new

        sudo_nuls = [
            "CO*01.-04",
            "ER*01.-03",
            "CO*M.01",
            "KEL*02.-14.1",
            "DI*02.15",
            "DI*02.16",
            "DI*02.11",
            "DI*02.12",
            "DI*02.22",
            "DI*02.09KEL*02M.05",
        ]
        # Di11/12 and 15/16 antithetical and same in ref so, yes, sudo null
        # (or more accurately, ref is third [unamed] ant)
        # DI*02.22/DI*02.09 (ant 9/22) seem to have a third antigen as well,
        #  ~300 examples in UKB - not a sudo null, but here till I find a better place
        # KEL*02M.05', is just KO, no other info....
        for ant in new_antigens:
            if (
                ant.antithetical_antigen
                and pair.allele1.genotype not in sudo_nuls
                and pair.allele2.genotype not in sudo_nuls
            ):  # TODO move to dif func
                final_no_expressed = count_expressed_ants(ant, base_names_new)
                to_neg = 'to_neg' in str(pair) and 'RHCE' in str(pair)
                if not to_neg: 
                    try:
                        assert final_no_expressed == 2
                    except AssertionError:
                        ic(
                        "Expressed antigens != 2!",
                        bg.sample,
                        str(pair),
                        ant,
                        final_no_expressed,
                        null
                    )

    for pair, merged_pheno in new_phenos:
        bg.phenotypes[ant_type][pair] = merged_pheno
 
    return bg


def make_ant(antithetical_ant, ant_type, bg_type, expressed):
    """Create an antigen instance from an antithetical antigen.

    The antigen class is chosen based on the blood group type and phenotype type.
    This function uses the base name of the provided antithetical antigen and sets
    the antigen as homozygous with the specified expressed state. Antithetical
    relationships are assigned from the global ANTITHETICAL dictionary.

    Args:
        antithetical_ant (an.Antigen): An antigen whose base name is used.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) of the antigen.
        bg_type (BgName): The blood group type to determine the antigen class.
        expressed (bool): Flag indicating whether the antigen is expressed.

    Returns:
        an.Antigen: A newly instantiated antigen object.
    """
    ant_class = choose_class_type(bg_type, ant_type)
    return ant_class(
        given_name=antithetical_ant.base_name,
        expressed=expressed,  # False if ant.expressed else True,
        homozygous=True,
        antithetical_relationships=ANTITHETICAL[ant_type].get(bg_type, {}),
    )


@apply_to_dict_values
def internal_anithetical_consistency_HOM(
    bg: BloodGroup, ant_type: PhenoType
) -> BloodGroup:
    """Ensure internal antithetical consistency for homozygous antigen pairs.

    This function enforces consistency among antigens for a BloodGroup by
    supplementing the existing phenotype mappings with antithetical antigen objects
    when necessary. For each allele pair in bg.phenotypes corresponding to the given
    phenotype type, if the allele pair is informative (i.e. its first allele's phenotype
    is not missing) and an antigen is homozygous with defined antithetical relationships,
    the function checks whether each antithetical antigen (based on its base name) is
    already present. If not, a new antigen is created using the `make_ant` helper function,
    with its expressed state set inversely to the original antigen's expressed flag.

    Example scenario:
        - Given a sample where KN*01.07/KN*01.07 is observed and KN*01.07 has a defined
          set of antigens such as KN:7, -9, 10.
        - The expected output may include a full set like:
          KN:1, -2, 3, -6, -4, 7, 5, 8, -9, 10, 11, -12, 13.
        - Explanation:
            * If KN7 is homozygous, then its antithetical antigen KN4 should not be present
              as a positive value, enforcing the relationship that 4 is antithetical to 7.
            * Similarly, if there is no supporting evidence for KN6 on the positive side,
              it remains negative.

    Args:
        bg (BloodGroup): The BloodGroup object containing phenotype data.
        ant_type (PhenoType): The phenotype type (e.g., numeric or alphanumeric) for which
            the antigen consistency should be enforced.

    Returns:
        BloodGroup: The updated BloodGroup with merged antigen lists that respect antithetical
            consistency for homozygous alleles.
    """
    
    new_phenos = []
    for pair, antigens in bg.phenotypes[ant_type].items():
        null = pair.allele1.null or pair.allele2.null
        new_antigens = []
        if pair.allele1.phenotype == ".":
            continue
        base_names = [ant2.base_name for ant2 in antigens]
        base_names_dict = {ant2.base_name: ant2 for ant2 in antigens}

        for ant in antigens:
            if ant.antithetical_antigen and ant.homozygous:
                if len(ant.antithetical_antigen) > 1:
                    no_expressed = count_expressed_ants(ant, base_names_dict)
                    if null and no_expressed != 0:
                        logger.warning("ensure 0 expressed for null !!!")
                    if not null and no_expressed != 2:
                        logger.warning("ensure 2 expressed !!!")
                for antithetical_ant in ant.antithetical_antigen:
                    if antithetical_ant.base_name not in base_names:
                        bg_type = BgName.from_string(bg.type)
                        # TODO freeze ants again??? - should never set to expressed?
                        new_antigens.append(
                            make_ant(
                                antithetical_ant,
                                ant_type,
                                bg_type,
                                False if ant.expressed else True,
                            )
                        )

        new_phenos.append((pair, new_antigens + antigens))
    for pair, merged_pheno in new_phenos:
        bg.phenotypes[ant_type][pair] = merged_pheno

    return bg


@apply_to_dict_values
def include_first_antithetical_pair(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Include the first antithetical pair from a reference into the blood group's
    phenotype.

    If the blood group is one of the FUT or ABO types, the function returns immediately
    with no changes. It then checks how many primary antitheticals are required for
    this blood group type. If zero, it returns immediately again. If no reference is
    found, it also returns immediately. Otherwise, it appends the required number of
    reference antigens to each pair in the phenotype if they are missing.

    Args:
        bg (BloodGroup): The blood group to update.
        ant_type (PhenoType): The domain of phenotypes to modify.

    Returns:
        BloodGroup: The updated (or unmodified) blood group.
    """
    if bg.type in ["FUT1", "FUT2", "FUT3", "ABO", "RHCE"]:
        return bg
    no_of_positions_required = number_of_primary_antitheticals.get(bg.type, 2)
    if no_of_positions_required == 0:
        return bg
    new_phenos = []
    ref = bg.misc[f"ref_{str(ant_type)}"]
    if ref is not None:
        # reference = ref.copy()
        reference = copy.deepcopy(ref)
    else:
        return bg
    for pair, merged_pheno in bg.phenotypes[ant_type].items():
        for i, name_ant in enumerate(reference.items(), start=1):
            _, ref_ant_list = name_ant
            assert len(ref_ant_list) == 1
            ref_ant = ref_ant_list[0]

            if ref_ant.base_name not in [
                allele_ant.base_name for allele_ant in merged_pheno
            ]:
                merged_pheno.append(ref_ant)
            if i == no_of_positions_required:
                break
        new_phenos.append((pair, merged_pheno))
    for pair, merged_pheno in new_phenos:
        bg.phenotypes[ant_type][pair] = merged_pheno

    return bg


@apply_to_dict_values
def sort_antigens(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Sort antigens for each allele pair in a BloodGroup based on a reference order.

    This function retrieves reference antigen data from bg.misc corresponding to the
    provided phenotype type. It then creates an ordering map from the reference keys
    (antigen base names) and sorts the merged antigen lists (bg.phenotypes) for each allele
    pair according to this order. Antigens with base names not found in the reference
    are placed at the end.

    Args:
        bg (BloodGroup): A BloodGroup object containing phenotype and miscellaneous data.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) for which antigens
            are being sorted.

    Returns:
        BloodGroup: The updated BloodGroup with sorted antigen lists in bg.phenotypes.
    """
    new_phenos = []
    reference = bg.misc[f"ref_{str(ant_type)}"]
    if reference is None:
        return bg
    else:
        ref = copy.deepcopy(reference)
    order_map = {name: position for position, name in list(enumerate(ref.keys()))}
    for pair, merged_pheno in bg.phenotypes[ant_type].items():
        sorted_merged_pheno = sorted(
            merged_pheno,
            key=lambda antigen: order_map.get(antigen.base_name, float("inf")),
        )
        new_phenos.append((pair, sorted_merged_pheno))
    for pair, sorted_merged_pheno in new_phenos:
        bg.phenotypes[ant_type][pair] = sorted_merged_pheno

    return bg


@apply_to_dict_values
def phenos_to_str(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Convert a list of antigen objects into a consolidated phenotype string.

    If the blood group has a RAW allele, its phenotype (before the first colon)
    is used as a prefix. If there's no RAW allele, the POS allele's prefix is used.
    Then, for each phenotype pair in ``bg.phenotypes[ant_type]``, a comma-separated
    string of antigen names is created. If ``bg.type`` is 'ABO', the list of antigens
    is sorted before joining. Otherwise, the order remains as is.

    Args:
        bg (BloodGroup): The blood group object containing alleles and phenotypes.
        ant_type (PhenoType): The phenotype category (e.g., alphanumeric or numeric).

    Returns:
        BloodGroup: The updated blood group with phenotype strings in
        ``bg.phenotypes[ant_type]`` instead of lists of antigen objects.

    Raises:
        IndexError: If neither RAW nor POS allele lists contain any alleles.
    """
    try:
        allele_name = bg.alleles[AlleleState.RAW][0].phenotype.split(":")[0]
    except IndexError:
        allele_name = bg.alleles[AlleleState.POS][0].phenotype.split(":")[0]
        logger.warning(f"Why doesnt this have RAW???? {bg.alleles}")
    for pair, merged_pheno in bg.phenotypes[ant_type].items():
        ants = [ant.name for ant in merged_pheno]
        as_str = ",".join(sorted(ants)) if bg.type == "ABO" else ",".join(ants)
        pheno = (
            as_str if ant_type == PhenoType.alphanumeric else f"{allele_name}:{as_str}"
        )
        bg.phenotypes[ant_type][pair] = pheno

    return bg


@apply_to_dict_values
def combine_anitheticals(bg: BloodGroup) -> BloodGroup:
    def split_list(input_list: list[str]) -> tuple[list[str], list[str]]:
        """
        Splits the input list into two lists based on specific criteria.

        Args:
            input_list (List[str]): The list of strings to be split.

        Returns:
            Tuple[List[str], List[str]]: A tuple containing two lists:
                - The first list contains strings that have a '(' and there are more
                than one string starting with the same prefix (substring before '(').
                - The second list contains all other strings.
        Example:
        Write a function that takes a list of strs:

        Ie
        Example1 = ['Kn(a+)','Kn(b-)','McC(a+)','Sl1+', 'Yk(a+)','McC(b-)', 'Vil-',
        'Sl3+','KCAM-','KDAS+','DACY-','YCAD+','KNMB+']
        Or
        Example2 = ['AUG1+', 'At(a+)', 'ATML-', 'ATAM+']

        And splits the list into 2 based on if the strs have a '(' and if there are
        more than 1 of the same type ('Kn(a+)' and 'Kn(b-)' are the same type because
        they both start with 'kn')

        The output from Example1 would be
        List_with_parens = ['Kn(a+)','Kn(b-)','McC(a+)','McC(b-)']
        Rest = ['Sl1+', 'Yk(a+)', 'Vil-','Sl3+','KCAM-','KDAS+','DACY-','YCAD+','KNMB+']

        Example2
        List_with_parens = []
        Rest = ['AUG1+', 'At(a+)', 'ATML-', 'ATAM+']
        """
        parens: list[str] = []
        rest: list[str] = []
        strings_with_parens: list[str] = [s for s in input_list if "(" in s]
        prefixes: list[str] = [s.split("(")[0] for s in strings_with_parens]
        prefix_counts: Counter = Counter(prefixes)
        for s in input_list:
            if "(" in s:
                prefix = s.split("(")[0]
                if prefix_counts[prefix] > 1:
                    parens.append(s)
                else:
                    rest.append(s)
            else:
                rest.append(s)

        return parens, rest

    to_update = []
    for pair, pheno in bg.phenotypes[PhenoType.alphanumeric].items():
        antigens = pheno.split(",")
        assert len(antigens) > 0

        has_paren, no_paren = split_list(antigens)
        if has_paren and no_paren:
            combined = ",".join([combine_expressions(has_paren)] + no_paren)
        elif has_paren:
            combined = ",".join([combine_expressions(has_paren)])
        elif no_paren:
            combined = ",".join(no_paren)

        to_update.append((pair, combined))

    for pair, combined in to_update:
        bg.phenotypes[PhenoType.alphanumeric][pair] = combined
    return bg


def combine_expressions(components: list[str]) -> str:
    """
    Combines multiple expressions with the same prefix into a single expression.

    This function takes a string containing multiple expressions separated by commas,
    where each expression has the same prefix (e.g., 'In', 'LW', 'Tc') followed by
    content in parentheses. It combines these expressions into a single expression
    with the same prefix and concatenated content.

    Args:
        expression (str): A string containing comma-separated expressions to be
        combined.

    Returns:
        str: A single combined expression.

    Examples:
        ['In(a-)','In(b+)'] > 'In(a-b+)'
        ['LW(a+)','LW(b-)'] > 'LW(a+b-)'
        ['Tc(a+)','Tc(b-)','Tc(c-)'] > 'Tc(a+b-c-)'
        ['Au(a-)', 'Lu(a-)', 'Au(b+)', 'Lu(b+)'] > 'Au(a-b+),Lu(a-b+)'

    """

    def extract_content(comp: str) -> str:
        """
        Extracts the content between parentheses from a component.

        Args:
            comp (str): A component string containing content in parentheses.

        Returns:
            str: The content between the parentheses.
        """
        start: int = comp.index("(") + 1
        end: int = comp.rindex(")")
        return comp[start:end]

    clustered = defaultdict(list)
    for phenotype in components:
        prefix_end: int = phenotype.index("(")
        prefix: str = phenotype[:prefix_end]
        clustered[prefix].append(phenotype)
    pheno = []
    for prefix, sub_components in clustered.items():
        combined_content: str = "".join(
            sorted(extract_content(comp) for comp in sub_components)
        )
        pheno.append(f"{prefix}({combined_content})")
    return ",".join(pheno)


def null_or_mod(pair: Pair, check: str) -> bool:
    """Check if the given pattern is present in both allele genotypes of a pair.

    This function constructs a pattern by appending a period ('.') to the provided
    check string, converts both allele genotypes to uppercase, and verifies whether the
    pattern exists in both.

    Args:
        pair (Pair): A Pair object containing two allele objects.
        check (str): A substring used to build the pattern to search for.

    Returns:
        bool: True if the pattern is present in both allele genotypes, otherwise False.
    """
    
    pattern = f"{check}." #TODO AUG*01N not covered!
    return (
        pattern in pair.allele1.genotype.upper()
        and pattern in pair.allele2.genotype.upper()
    )


@apply_to_dict_values
def modify_FY(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Modify FY BloodGroup phenotypes to indicate erythroid cell restriction.

    For BloodGroup of type "FY", this function appends a suffix to the phenotype string
    if either allele's genotype is one of the specified erythroid null alleles and if both
    alleles contain the 'N.' pattern. This indicates that the phenotype is restricted to
    erythroid cells only.

    Args:
        bg (BloodGroup): The BloodGroup object to be modified.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) determining
            which phenotype mapping to modify.

    Returns:
        BloodGroup: The updated BloodGroup with modified FY phenotypes.
    """
    if bg.type != "FY":
        return bg
    for pair, pheno in bg.phenotypes[ant_type].items():
        erythroid = ["FY*02N.01", "FY*01N.08", "FY*01N.01"]
        if (
            pair.allele1.genotype in erythroid or pair.allele2.genotype in erythroid
        ) and null_or_mod(pair, "N"):
            # only if 2 nulls
            # TODO think about how to handle nulls better
            pheno += "_erythroid_cells_only"
        bg.phenotypes[ant_type][pair] = pheno

    return bg

@apply_to_dict_values
def modify_FY2(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """
    add FYX to FY*02W.01 and FY*02W.02

    Args:
        bg (BloodGroup): The BloodGroup object to be modified.
        ant_type (PhenoType): The phenotype type (alphanumeric) whose
            corresponding phenotype mapping is to be updated.

    Returns:
        BloodGroup: The updated BloodGroup with modified FY phenotypes.
    """
    if bg.type != "FY":
        return bg
    for pair, pheno in bg.phenotypes[ant_type].items():
        FYX = ["FY*02W.01", "FY*02W.02"]

        if (
            pair.allele1.genotype in FYX or pair.allele2.genotype in FYX
        ) and "b+w)" in pheno:
            pheno += ',Fyx'

        bg.phenotypes[ant_type][pair] = pheno

    return bg

@apply_to_dict_values
def modify_KEL(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Modify KEL BloodGroup phenotypes based on allele null/modification patterns.

    For BloodGroup of type "KEL", this function updates the phenotype mapping in
    bg.phenotypes for the provided phenotype type. If an allele pair matches the
    'M.' pattern (as determined by null_or_mod), the phenotype is set to "Kmod".
    Similarly, if an allele pair matches the 'N.' pattern, the phenotype is set to "KO".

    Args:
        bg (BloodGroup): The BloodGroup object to be modified.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) whose
            corresponding phenotype mapping is to be updated.

    Returns:
        BloodGroup: The updated BloodGroup with modified KEL phenotypes.
    """
    if bg.type != "KEL":
        return bg
    for pair in bg.phenotypes[ant_type]:
        if null_or_mod(pair, "M"):
            bg.phenotypes[ant_type][pair] = "Kmod"
    for pair in bg.phenotypes[ant_type]:
        if null_or_mod(pair, "N"):
            bg.phenotypes[ant_type][pair] = "KO"

    return bg

@apply_to_dict_values
def modify_CROM(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Modify CROM BloodGroup phenotypes based on allele null/modification patterns.

    For BloodGroup of type "CROM", this function updates the phenotype mapping in
    bg.phenotypes for the provided phenotype type. 
    if an allele pair matches the 'N.' pattern, the phenotype is changed from
    IFC- to Inab

    Args:
        bg (BloodGroup): The BloodGroup object to be modified.
        ant_type (PhenoType): The phenotype type (numeric or alphanumeric) whose
            corresponding phenotype mapping is to be updated.

    Returns:
        BloodGroup: The updated BloodGroup with modified CROM phenotypes.
    """
    if bg.type != "CROM":
        return bg
    
    for pair in bg.phenotypes[ant_type]:
        if null_or_mod(pair, "N"):
            bg.phenotypes[ant_type][pair] = "Inab"

    return bg


@apply_to_dict_values
def modify_MNS(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """
    U+w -> U+var

    Args:
        bg (BloodGroup): The BloodGroup object to be modified.
        ant_type (PhenoType): The phenotype type (alphanumeric) whose
            corresponding phenotype mapping is to be updated.

    Returns:
        BloodGroup: The updated BloodGroup with modified MNS (GYPB) phenotypes.
    """
    if bg.type != "GYPB":
        return bg
    for pair, pheno in bg.phenotypes[ant_type].items():
        weak_mod = "GYPB*04.03"
        var_mod = [
            "GYPB*04N.03",
            "GYPB*04N.01",
            "GYPB*03N.01",
            "GYPB*03N.02",
            "GYPB*03N.03",
            "GYPB*03N.04",
            "GYPB*03N.05",
            "GYPB*03N.06",
            "GYPB*03N.07",
        ]
        if (
            pair.allele1.genotype == weak_mod or pair.allele2.genotype == weak_mod
        ) and (pair.allele1.genotype in var_mod or pair.allele2.genotype in var_mod):
            pheno += "var"
        elif pair.allele1.genotype in var_mod or pair.allele2.genotype in var_mod:
            pheno = pheno.replace("U+w", "U+var")

        bg.phenotypes[ant_type][pair] = pheno

    return bg





@apply_to_dict_values
def re_order_KEL(bg: BloodGroup, ant_type: PhenoType) -> BloodGroup:
    """Reorder KEL phenotypes so that K+ and K- appear first.

    This function processes phenotype strings for BloodGroup objects of type "KEL". It
    reorders the antigens in the phenotype string such that antigens starting with "K+"
    or "K-" are placed at the beginning, followed by the remaining antigens. For example,
    a phenotype string "Js(a+b+),K-,k+" will be rearranged to "K-,k+,Js(a+b+)".

    Args:
        bg (BloodGroup): The BloodGroup object containing phenotype mappings.
        ant_type (PhenoType): The phenotype type (e.g., numeric or alphanumeric) to be
            processed.

    Returns:
        BloodGroup: The updated BloodGroup with re-ordered KEL phenotypes.
    """
    if bg.type != "KEL":
        return bg

    big_k = ("K+", "K-")

    def sorter(unsorted_pheno):
        sorted_antigens = []
        for i in range(2):
            for ant in unsorted_pheno.strip().split(","):
                if (
                    i == 0
                    and ant.upper().startswith(big_k)
                    or i == 1
                    and not ant.upper().startswith(big_k)
                ):
                    sorted_antigens.append(ant)

        return ",".join(sorted_antigens)

    for pair, pheno in bg.phenotypes[ant_type].items():
        if not pheno.upper().startswith(big_k):
            pheno = sorter(pheno)
        bg.phenotypes[ant_type][pair] = pheno

    return bg


numeric_calculators = {}

number_of_primary_antitheticals = {
    "RHAG": 1,
    "JMH": 1,
    "OK": 1,
    "KN": 13,
    "CROM": 1,
    "LW": 1,
    "Vel": 1,
    "LU": 0,
}



import unittest
from collections import defaultdict
from unittest.mock import MagicMock, patch

# --- Import the actual components from your project ---
from rbceq2.core_logic.alleles import Allele, BloodGroup, Pair
from rbceq2.core_logic.constants import AlleleState
from rbceq2.core_logic.utils import Zygosity
# Import the functions to be tested
from rbceq2.filters.knops import (
    ensure_co_existing_HET_SNP_used,
    filter_co_existing_in_other_allele,
    filter_co_existing_pairs,
    filter_co_existing_subsets,
    filter_co_existing_with_normal,
    filter_coexisting_pairs_on_antithetical_zygosity,
    parse_bio_info2,
    remove_unphased_co,
)
# This is a dependency of the functions under test
from rbceq2.filters.shared_filter_functionality import flatten_alleles


# --- Helper Mock Class for Testing ---
class MockAllele(Allele):
    def __init__(self, **kwargs):
        defaults = {
            "phenotype": ".", "genotype_alt": ".", "phenotype_alt": ".",
            "defining_variants": frozenset(), "null": False, "weight_geno": 1000,
            "reference": False, "sub_type": "default",
        }
        defaults.update(kwargs)
        super().__init__(**defaults)

    def __eq__(self, other):
        return isinstance(other, Allele) and self.genotype == other.genotype

    def __hash__(self):
        return hash(self.genotype)

class MockPair:
    def __init__(self, allele1, allele2):
        self.alleles = (allele1, allele2)
        self.allele1 = allele1
        self.allele2 = allele2
    @property
    def comparable(self): return [frozenset(a.genotype.split("+")) for a in self.alleles]
    @property
    def contains_reference(self): return any(a.reference for a in self.alleles)
    @property
    def same_subtype(self): return self.allele1.sub_type == self.allele2.sub_type
    def __iter__(self): return iter(self.alleles)
    def __eq__(self, other): return isinstance(other, MockPair) and self.alleles == other.alleles
    def __hash__(self): return hash(self.alleles)
    def __repr__(self): return f"Pair(Genotype: {self.allele1.genotype}/{self.allele2.genotype})"


# --- Final, Consolidated, and Complete Test Class ---
class TestKnopsCoExistingFilters(unittest.TestCase):
    def setUp(self):
        """Set up a fresh mock BloodGroup object for each test."""
        self.mock_bg = MagicMock(spec=BloodGroup)
        self.mock_bg.alleles = {
            AlleleState.NORMAL: [], AlleleState.CO: [], AlleleState.RAW: [], AlleleState.FILT: []
        }
        self.mock_bg.filtered_out = defaultdict(list)
        self.mock_bg.variant_pool = {}
        self.mock_bg.variant_phase_set = {}
        self.mock_bg.type = "KN"

        def mock_remove(items_to_remove, filter_name, state=AlleleState.CO):
            self.mock_bg.filtered_out[filter_name].extend(items_to_remove)
            if self.mock_bg.alleles.get(state):
                self.mock_bg.alleles[state] = [item for item in self.mock_bg.alleles[state] if item not in items_to_remove]
        
        self.mock_bg.remove_pairs.side_effect = mock_remove
        self.mock_bg.remove_alleles.side_effect = mock_remove

    # --- Tests for filter_co_existing_with_normal ---
    def test_filter_with_normal_removes_pair_not_in_normal_list(self):
        pair_to_filter = MockPair(MockAllele(genotype="A"), MockAllele(genotype="B"))
        pair_to_keep = MockPair(MockAllele(genotype="C"), MockAllele(genotype="D"))
        self.mock_bg.alleles[AlleleState.NORMAL] = [pair_to_keep]
        self.mock_bg.alleles[AlleleState.CO] = [pair_to_filter, pair_to_keep]
        self.mock_bg.filtered_out[AlleleState.NORMAL] = [pair_to_filter]
        filter_co_existing_with_normal({1: self.mock_bg})
        self.mock_bg.remove_pairs.assert_called_once_with([pair_to_filter], "filter_co_existing_with_normal", AlleleState.CO)

    # --- Tests for parse_bio_info2 ---
    def test_parse_bio_info2(self):
        a1, a2, a3 = MockAllele(genotype="A+B"), MockAllele(genotype="C"), MockAllele(genotype="D")
        result = parse_bio_info2([MockPair(a1, a2), MockPair(a2, a3)])
        self.assertEqual(result, [[frozenset({"A", "B"}), frozenset({"C"})], [frozenset({"C"}), frozenset({"D"})]])

    # --- Tests for filter_co_existing_subsets ---
    def test_filter_subsets_removes_subset_pair(self):
        # CORRECTED TEST DATA
        aA, aB, aC = MockAllele(genotype="A"), MockAllele(genotype="B"), MockAllele(genotype="C")
        aAB = MockAllele(genotype="A+B")
        
        # This pair's flattened alleles are {'A', 'B'}
        pair_subset = MockPair(aA, aB)
        # This pair's flattened alleles are {'A', 'B', 'C'}, a proper superset
        pair_superset = MockPair(aAB, aC)
        
        self.mock_bg.alleles[AlleleState.CO] = [pair_subset, pair_superset]
        self.mock_bg.alleles[AlleleState.RAW] = [aA, aB, aC, aAB]
        self.mock_bg.variant_pool = {}
        
        filter_co_existing_subsets({1: self.mock_bg})
        
        self.mock_bg.remove_pairs.assert_called_once_with([pair_subset], "filter_co_existing_subsets", AlleleState.CO)

    def test_filter_subsets_does_not_remove_permutations(self):
        aA, aB = MockAllele(genotype="A"), MockAllele(genotype="B")
        pair1, pair2 = MockPair(aA, aB), MockPair(aB, aA)
        self.mock_bg.alleles[AlleleState.CO] = [pair1, pair2]
        self.mock_bg.alleles[AlleleState.RAW] = [aA, aB]
        filter_co_existing_subsets({1: self.mock_bg})
        self.mock_bg.remove_pairs.assert_not_called()

    # --- Tests for filter_coexisting_pairs_on_antithetical_zygosity ---
    def test_co_antithetical_removes_mismatched_subtypes(self):
        self.mock_bg.type = "SystemX"
        antitheticals = {"SystemX": ["S1", "S2"]}
        a_s1, a_s1_alt = MockAllele(genotype="A", sub_type="S1"), MockAllele(genotype="B", sub_type="S1")
        a_s2 = MockAllele(genotype="C", sub_type="S2")
        self.mock_bg.alleles[AlleleState.NORMAL] = [MockPair(a_s1, a_s2)]
        pair_to_remove = MockPair(a_s1, a_s1_alt)
        self.mock_bg.alleles[AlleleState.CO] = [pair_to_remove]
        filter_coexisting_pairs_on_antithetical_zygosity({1: self.mock_bg}, antitheticals)
        self.mock_bg.remove_pairs.assert_called_once_with([pair_to_remove], "filter_co_pairs_on_antithetical_zygosity", AlleleState.CO)

    # --- Tests for ensure_co_existing_HET_SNP_used ---
    @patch("rbceq2.filters.knops.check_var", return_value=2)
    def test_ensure_co_het_removes_pair_with_multiple_hits(self, mock_check_var):
        pair_ambiguous = MockPair(MockAllele(genotype="A"), MockAllele(genotype="B"))
        self.mock_bg.type = "KN"
        self.mock_bg.variant_pool = {"het_var": Zygosity.HET}
        self.mock_bg.alleles[AlleleState.CO] = [pair_ambiguous]
        ensure_co_existing_HET_SNP_used({1: self.mock_bg})
        mock_check_var.assert_called_with(self.mock_bg, pair_ambiguous, AlleleState.CO, "het_var")
        self.mock_bg.remove_pairs.assert_called_once_with([pair_ambiguous], "ensure_co_existing_HET_SNP_used", AlleleState.CO)

    # --- Tests for remove_unphased_co ---
    @patch("rbceq2.filters.knops.flatten_alleles")
    @patch("rbceq2.filters.knops.identify_unphased")
    def test_remove_unphased_co_removes_unphased_pair(self, mock_identify, mock_flatten):
        unphased_pair = MockPair(MockAllele(genotype="A"), MockAllele(genotype="B"))
        mock_identify.return_value = [unphased_pair.allele1]
        self.mock_bg.alleles[AlleleState.CO] = [unphased_pair]
        self.mock_bg.type = "KN"
        remove_unphased_co({1: self.mock_bg}, phased=True)
        self.mock_bg.remove_pairs.assert_called_once_with([unphased_pair], "remove_unphased_co", AlleleState.CO)

    # --- Tests for filter_co_existing_pairs ---
    def test_filter_co_pairs_removes_mushed_with_slash_in_subtype(self):
        mushed_pair = MockPair(MockAllele(genotype="A"), MockAllele(genotype="B+C", sub_type="KN*01/KN*02", genotype_alt="mushed"))
        self.mock_bg.alleles[AlleleState.CO] = [mushed_pair]
        filter_co_existing_pairs({1: self.mock_bg})
        self.mock_bg.remove_pairs.assert_called_once_with([mushed_pair], "filter_co_existing_pairs", AlleleState.CO)

    # --- Tests for filter_co_existing_in_other_allele ---
    def test_filter_co_in_other_removes_mushed_matching_real_allele_variants(self):
        mushed_allele = MockAllele(genotype="A+B", defining_variants={"v1", "v2"}, genotype_alt="mushed")
        real_allele = MockAllele(genotype="C_REAL", defining_variants={"v1", "v2"})
        pair_to_remove = MockPair(mushed_allele, MockAllele(genotype="D", defining_variants={"v3"}))
        self.mock_bg.alleles[AlleleState.CO] = [pair_to_remove]
        self.mock_bg.alleles[AlleleState.FILT] = [real_allele]
        filter_co_existing_in_other_allele({1: self.mock_bg})
        self.mock_bg.remove_pairs.assert_called_once_with([pair_to_remove], "filter_co_existing_in_other_allele", AlleleState.CO)


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)
    
# import unittest
# from collections import defaultdict
# from unittest.mock import MagicMock, patch

# # --- Import the actual components from your project ---
# from rbceq2.core_logic.alleles import Allele, BloodGroup
# from rbceq2.core_logic.constants import AlleleState
# from rbceq2.core_logic.utils import Zygosity

# # Import the functions to be tested
# from rbceq2.filters.knops import (
#     ensure_co_existing_HET_SNP_used,
#     filter_co_existing_in_other_allele,
#     filter_co_existing_pairs,
#     filter_co_existing_subsets,
#     filter_co_existing_with_normal,
#     filter_coexisting_pairs_on_antithetical_zygosity,
#     parse_bio_info2,
#     remove_unphased_co,
# )
# # This is a dependency of the functions under test


# # --- Helper Mock Class for Testing ---
# class MockAllele(Allele):
#     def __init__(self, **kwargs):
#         defaults = {
#             "phenotype": ".",
#             "genotype_alt": ".",
#             "phenotype_alt": ".",
#             "defining_variants": frozenset(),
#             "null": False,
#             "weight_geno": 1000,
#             "reference": False,
#             "sub_type": "default",
#         }
#         defaults.update(kwargs)
#         super().__init__(**defaults)

#     def __eq__(self, other):
#         return isinstance(other, Allele) and self.genotype == other.genotype

#     def __hash__(self):
#         return hash(self.genotype)


# class MockPair:
#     def __init__(self, allele1, allele2):
#         self.alleles = (allele1, allele2)
#         self.allele1 = allele1
#         self.allele2 = allele2

#     @property
#     def comparable(self):
#         return [frozenset(a.genotype.split("+")) for a in self.alleles]

#     @property
#     def contains_reference(self):
#         return any(a.reference for a in self.alleles)

#     @property
#     def same_subtype(self):
#         return self.allele1.sub_type == self.allele2.sub_type

#     def __iter__(self):
#         return iter(self.alleles)

#     def __eq__(self, other):
#         return isinstance(other, MockPair) and self.alleles == other.alleles

#     def __hash__(self):
#         return hash(self.alleles)

#     def __repr__(self):
#         return f"Pair(Genotype: {self.allele1.genotype}/{self.allele2.genotype})"


# # --- Final, Consolidated, and Complete Test Class ---
# class TestKnopsCoExistingFilters(unittest.TestCase):
#     def setUp(self):
#         """Set up a fresh mock BloodGroup object for each test."""
#         self.mock_bg = MagicMock(spec=BloodGroup)
#         self.mock_bg.alleles = {
#             AlleleState.NORMAL: [],
#             AlleleState.CO: [],
#             AlleleState.RAW: [],
#             AlleleState.FILT: [],
#         }
#         self.mock_bg.filtered_out = defaultdict(list)
#         self.mock_bg.variant_pool = {}
#         self.mock_bg.variant_phase_set = {}
#         self.mock_bg.type = "KN"

#         def mock_remove(items_to_remove, filter_name, state=AlleleState.CO):
#             self.mock_bg.filtered_out[filter_name].extend(items_to_remove)
#             if self.mock_bg.alleles.get(state):
#                 self.mock_bg.alleles[state] = [
#                     item
#                     for item in self.mock_bg.alleles[state]
#                     if item not in items_to_remove
#                 ]

#         self.mock_bg.remove_pairs.side_effect = mock_remove
#         self.mock_bg.remove_alleles.side_effect = mock_remove

#     # --- Tests for filter_co_existing_with_normal ---
#     def test_filter_with_normal_removes_pair_not_in_normal_list(self):
#         pair_to_filter = MockPair(MockAllele(genotype="A"), MockAllele(genotype="B"))
#         pair_to_keep = MockPair(MockAllele(genotype="C"), MockAllele(genotype="D"))
#         self.mock_bg.alleles[AlleleState.NORMAL] = [pair_to_keep]
#         self.mock_bg.alleles[AlleleState.CO] = [pair_to_filter, pair_to_keep]
#         self.mock_bg.filtered_out[AlleleState.NORMAL] = [pair_to_filter]
#         filter_co_existing_with_normal({1: self.mock_bg})
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_to_filter], "filter_co_existing_with_normal", AlleleState.CO
#         )

#     def test_filter_with_normal_skips_mushed_alleles(self):
#         mushed_allele = MockAllele(genotype="A", genotype_alt="mushed")
#         pair_to_skip = MockPair(mushed_allele, MockAllele(genotype="B"))
#         self.mock_bg.alleles[AlleleState.NORMAL] = []
#         self.mock_bg.alleles[AlleleState.CO] = [pair_to_skip]
#         filter_co_existing_with_normal({1: self.mock_bg})
#         self.mock_bg.remove_pairs.assert_not_called()

#     # --- Tests for parse_bio_info2 ---
#     def test_parse_bio_info2(self):
#         a1, a2, a3 = (
#             MockAllele(genotype="A+B"),
#             MockAllele(genotype="C"),
#             MockAllele(genotype="D"),
#         )
#         result = parse_bio_info2([MockPair(a1, a2), MockPair(a2, a3)])
#         self.assertEqual(
#             result,
#             [
#                 [frozenset({"A", "B"}), frozenset({"C"})],
#                 [frozenset({"C"}), frozenset({"D"})],
#             ],
#         )

#     # --- Tests for filter_co_existing_subsets ---
#     def test_filter_subsets_removes_subset_pair(self):
#         aA, aB, aAB = (
#             MockAllele(genotype="A", defining_variants={"vA"}),
#             MockAllele(genotype="B", defining_variants={"vB"}),
#             MockAllele(genotype="A+B", defining_variants={"vA", "vB"}),
#         )
#         pair_subset, pair_superset = MockPair(aA, aB), MockPair(aAB, aB)
#         self.mock_bg.alleles[AlleleState.CO] = [pair_subset, pair_superset]
#         self.mock_bg.alleles[AlleleState.RAW] = [aA, aB, aAB]
#         self.mock_bg.variant_pool = {"vA": Zygosity.HOM, "vB": Zygosity.HOM}
#         filter_co_existing_subsets({1: self.mock_bg})
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_subset], "filter_co_existing_subsets", AlleleState.CO
#         )

#     def test_filter_subsets_does_not_remove_permutations(self):
#         aA, aB = (
#             MockAllele(genotype="A", defining_variants={"vA"}),
#             MockAllele(genotype="B", defining_variants={"vB"}),
#         )
#         pair1, pair2 = MockPair(aA, aB), MockPair(aB, aA)
#         self.mock_bg.alleles[AlleleState.CO] = [pair1, pair2]
#         self.mock_bg.alleles[AlleleState.RAW] = [aA, aB]
#         self.mock_bg.variant_pool = {"vA": Zygosity.HOM, "vB": Zygosity.HOM}
#         filter_co_existing_subsets({1: self.mock_bg})
#         self.mock_bg.remove_pairs.assert_not_called()

#     # --- Tests for filter_coexisting_pairs_on_antithetical_zygosity ---
#     def test_co_antithetical_removes_mismatched_subtypes(self):
#         self.mock_bg.type = "SystemX"
#         antitheticals = {"SystemX": ["S1", "S2"]}
#         a_s1, a_s1_alt = (
#             MockAllele(genotype="A", sub_type="S1"),
#             MockAllele(genotype="B", sub_type="S1"),
#         )
#         a_s2 = MockAllele(genotype="C", sub_type="S2")
#         self.mock_bg.alleles[AlleleState.NORMAL] = [MockPair(a_s1, a_s2)]
#         pair_to_remove = MockPair(a_s1, a_s1_alt)
#         self.mock_bg.alleles[AlleleState.CO] = [pair_to_remove]
#         filter_coexisting_pairs_on_antithetical_zygosity(
#             {1: self.mock_bg}, antitheticals
#         )
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_to_remove], "filter_co_pairs_on_antithetical_zygosity", AlleleState.CO
#         )

#     # --- Tests for ensure_co_existing_HET_SNP_used ---
#     @patch("rbceq2.filters.knops.check_var")
#     def test_ensure_co_het_removes_pair_with_multiple_hits(self, mock_check_var):
#         mock_check_var.return_value = 2
#         pair_ambiguous = MockPair(
#             MockAllele(genotype="A", defining_variants={"v1"}),
#             MockAllele(genotype="B", defining_variants={"v2"}),
#         )
#         self.mock_bg.type = "KN"
#         self.mock_bg.variant_pool = {"het_var": Zygosity.HET}
#         self.mock_bg.alleles[AlleleState.CO] = [pair_ambiguous]
#         ensure_co_existing_HET_SNP_used({1: self.mock_bg})
#         mock_check_var.assert_called_with(
#             self.mock_bg, pair_ambiguous, AlleleState.CO, "het_var"
#         )
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_ambiguous], "ensure_co_existing_HET_SNP_used", AlleleState.CO
#         )

#     # --- Tests for remove_unphased_co ---
#     @patch("rbceq2.filters.knops.flatten_alleles")
#     @patch("rbceq2.filters.knops.identify_unphased")
#     def test_remove_unphased_co_removes_unphased_pair(
#         self, mock_identify, mock_flatten
#     ):
#         unphased_pair = MockPair(
#             MockAllele(genotype="A", defining_variants={"vA"}),
#             MockAllele(genotype="B", defining_variants={"vB"}),
#         )
#         mock_identify.return_value = [unphased_pair.allele1]
#         self.mock_bg.alleles[AlleleState.CO] = [unphased_pair]
#         self.mock_bg.type = "KN"
#         remove_unphased_co({1: self.mock_bg}, phased=True)
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [unphased_pair], "remove_unphased_co", AlleleState.CO
#         )

#     # --- Tests for filter_co_existing_pairs ---
#     def test_filter_co_pairs_removes_mushed_with_slash_in_subtype(self):
#         mushed_pair = MockPair(
#             MockAllele(genotype="A"),
#             MockAllele(genotype="B+C", sub_type="KN*01/KN*02", genotype_alt="mushed"),
#         )
#         self.mock_bg.alleles[AlleleState.CO] = [mushed_pair]
#         filter_co_existing_pairs({1: self.mock_bg})
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [mushed_pair], "filter_co_existing_pairs", AlleleState.CO
#         )

#     # --- Tests for filter_co_existing_in_other_allele ---
#     def test_filter_co_in_other_removes_mushed_matching_real_allele_variants(self):
#         mushed_allele = MockAllele(
#             genotype="A+B", defining_variants={"v1", "v2"}, genotype_alt="mushed"
#         )
#         real_allele = MockAllele(genotype="C_REAL", defining_variants={"v1", "v2"})
#         pair_to_remove = MockPair(
#             mushed_allele, MockAllele(genotype="D", defining_variants={"v3"})
#         )
#         self.mock_bg.alleles[AlleleState.CO] = [pair_to_remove]
#         self.mock_bg.alleles[AlleleState.FILT] = [real_allele]
#         filter_co_existing_in_other_allele({1: self.mock_bg})
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_to_remove], "filter_co_existing_in_other_allele", AlleleState.CO
#         )


# if __name__ == "__main__":
#     unittest.main(argv=["first-arg-is-ignored"], exit=False)

#TODO check all this is covered b4 del
# import unittest
# from unittest.mock import MagicMock
# from rbceq2.core_logic.alleles import Allele, BloodGroup, Pair

# from rbceq2.core_logic.constants import AlleleState
# from rbceq2.core_logic.utils import Zygosity
# from rbceq2.filters.knops import (
#     ensure_co_existing_HET_SNP_used,
#     filter_co_existing_in_other_allele,
#     filter_co_existing_pairs,
#     filter_co_existing_subsets,
#     filter_co_existing_with_normal,
#     filter_coexisting_pairs_on_antithetical_zygosity,
#     parse_bio_info2,
#     remove_unphased_co,
# )

# from collections import defaultdict


# # --- Mock Data Classes (retained for test isolation and simplicity) ---
# # Using these mock classes allows us to test the filter logic without
# # depending on the internal implementation of the real Allele/Pair classes.
# class MockAllele:
#     def __init__(
#         self,
#         genotype,
#         defining_variants,
#         sub_type=".",
#         genotype_alt=".",
#         reference=False,
#     ):
#         self.genotype = genotype
#         self.defining_variants = frozenset(defining_variants)
#         self.sub_type = sub_type
#         self.genotype_alt = genotype_alt
#         self.reference = reference

#     def __repr__(self):
#         return f"Allele(genotype='{self.genotype}')"

#     def __eq__(self, other):
#         return isinstance(other, MockAllele) and self.genotype == other.genotype

#     def __hash__(self):
#         return hash(self.genotype)


# class MockPair:
#     def __init__(self, allele1, allele2):
#         self.alleles = (allele1, allele2)
#         self.allele1 = allele1
#         self.allele2 = allele2

#     @property
#     def comparable(self):
#         return [frozenset(a.genotype.split("+")) for a in self.alleles]

#     @property
#     def contains_reference(self):
#         return any(a.reference for a in self.alleles)

#     @property
#     def same_subtype(self):
#         return self.allele1.sub_type == self.allele2.sub_type

#     def __iter__(self):
#         return iter(self.alleles)

#     def __eq__(self, other):
#         return isinstance(other, MockPair) and self.alleles == other.alleles

#     def __hash__(self):
#         return hash(self.alleles)

#     def __repr__(self):
#         return f"Pair(Genotype: {self.allele1.genotype}/{self.allele2.genotype})"


# # --- Final, Corrected Test Class ---
# class TestCoExistingFilters(unittest.TestCase):
#     def setUp(self):
#         """Set up a fresh mock BloodGroup object for each test."""
#         self.mock_bg = MagicMock(spec=BloodGroup)
#         self.mock_bg.alleles = {
#             AlleleState.NORMAL: [],
#             AlleleState.CO: [],
#             AlleleState.RAW: [],
#             AlleleState.FILT: [],
#         }
#         self.mock_bg.filtered_out = defaultdict(list)
#         self.mock_bg.variant_pool = {}
#         self.mock_bg.variant_phase_set = {}
#         self.mock_bg.type = "KN"

#     def test_filter_co_existing_with_normal(self):
#         """Test filtering a CO pair if its equivalent NORMAL pair was already filtered."""
#         allele1 = MockAllele("KN*01.06", {"varA"})
#         allele2 = MockAllele("KN*01.07", {"varB"})
#         pair_to_filter = MockPair(allele1, allele2)

#         # Setup: The pair exists in both states but is marked as filtered in NORMAL
#         self.mock_bg.alleles[AlleleState.NORMAL] = []  # It was removed from here
#         self.mock_bg.alleles[AlleleState.CO] = [pair_to_filter]
#         self.mock_bg.filtered_out[AlleleState.NORMAL] = [pair_to_filter]

#         # Action: Call the decorated function with a dictionary
#         filter_co_existing_with_normal({1: self.mock_bg})

#         # Assert: remove_pairs should be called on the CO state
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_to_filter], "filter_co_existing_with_normal", AlleleState.CO
#         )

#     def test_parse_bio_info2(self):
#         """Test parsing of pairs into a list of frozensets (not decorated)."""
#         allele1 = MockAllele("A+B", {})
#         allele2 = MockAllele("C", {})
#         allele3 = MockAllele("D", {})
#         pair1 = MockPair(allele1, allele2)
#         pair2 = MockPair(allele2, allele3)

#         # Action: Call directly as it's not decorated
#         result = parse_bio_info2([pair1, pair2])

#         expected = [
#             [frozenset({"A", "B"}), frozenset({"C"})],
#             [frozenset({"C"}), frozenset({"D"})],
#         ]
#         self.assertEqual(result, expected)

#     def test_filter_co_existing_subsets(self):
#         """Test filtering of co-existing pairs that are subsets of others."""
#         allele_A = MockAllele("A", {"varA"})
#         allele_B = MockAllele("B", {"varB"})
#         allele_AB = MockAllele("A+B", {"varA", "varB"})
#         pair_subset = MockPair(allele_A, allele_B)
#         pair_superset = MockPair(allele_AB, allele_B)

#         self.mock_bg.alleles[AlleleState.CO] = [pair_subset, pair_superset]
#         self.mock_bg.alleles[AlleleState.RAW] = [allele_A, allele_B, allele_AB]
#         self.mock_bg.variant_pool = {"varA": Zygosity.HOM, "varB": Zygosity.HOM}

#         filter_co_existing_subsets({1: self.mock_bg})

#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_subset], "filter_co_existing_subsets", AlleleState.CO
#         )

#     def test_filter_coexisting_pairs_on_antithetical_zygosity(self):
#         """Test filtering based on antithetical zygosity rules."""
#         antitheticals = {"KN": ["KN*01", "KN*02"]}
#         allele_01 = MockAllele("A", {}, sub_type="KN*01")
#         allele_02 = MockAllele("B", {}, sub_type="KN*02")
#         allele_01_mut = MockAllele("C", {}, sub_type="KN*01")

#         valid_pair = MockPair(allele_01, allele_02)
#         invalid_pair = MockPair(allele_01, allele_01_mut)

#         # Setup: NORMAL alleles must have >1 subtype to trigger the filter logic.
#         # CO alleles contain the pair that should be filtered.
#         self.mock_bg.alleles[AlleleState.NORMAL] = [valid_pair]
#         self.mock_bg.alleles[AlleleState.CO] = [valid_pair, invalid_pair]

#         filter_coexisting_pairs_on_antithetical_zygosity(
#             {1: self.mock_bg}, antitheticals
#         )

#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [invalid_pair], "filter_co_pairs_on_antithetical_zygosity", AlleleState.CO
#         )

#     def test_ensure_co_existing_HET_SNP_used(self):
#         """Test removal of pairs where a HET SNP could ambiguously form alleles."""
#         allele_A = MockAllele("A", {"var1"})
#         allele_B = MockAllele("B", {"var2"})
#         pair_ambiguous = MockPair(allele_A, allele_B)

#         self.mock_bg.variant_pool = {"het_var": Zygosity.HET}
#         self.mock_bg.alleles[AlleleState.CO] = [pair_ambiguous]

#         # Patch check_var where it's defined and make it return > 1
#         with unittest.mock.patch(
#             "rbceq2.filters.shared_filter_functionality.check_var", return_value=2
#         ) as mock_check_var:
#             ensure_co_existing_HET_SNP_used({1: self.mock_bg})

#         mock_check_var.assert_called_with(
#             self.mock_bg, pair_ambiguous, AlleleState.CO, "het_var"
#         )
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_ambiguous], "ensure_co_existing_HET_SNP_used", AlleleState.CO
#         )

#     def test_remove_unphased_co(self):
#         """Test removal of co-existing pairs that are not correctly phased."""
#         allele1 = MockAllele("A", {"varA"})
#         allele2 = MockAllele("B", {"varB"})
#         unphased_pair = MockPair(allele1, allele2)

#         self.mock_bg.alleles[AlleleState.CO] = [unphased_pair]
#         self.mock_bg.type = "KN"  # Required for the function to proceed

#         # Patch identify_unphased where it's defined.
#         with unittest.mock.patch(
#             "rbceq2.filters.shared_filter_functionality.identify_unphased",
#             return_value=[allele1],
#         ) as mock_identify:
#             remove_unphased_co({1: self.mock_bg}, phased=True)

#         # Assert
#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [unphased_pair], "remove_unphased_co", AlleleState.CO
#         )

#     def test_filter_co_existing_pairs(self):
#         """Test removal of pairs with invalid 'mushed' alleles containing a '/' in subtype."""
#         allele_good = MockAllele("A", {"varA"})
#         allele_mushed = MockAllele(
#             "B+C", {"varB", "varC"}, sub_type="KN*01/KN*02", genotype_alt="mushed"
#         )
#         pair_to_remove = MockPair(allele_good, allele_mushed)

#         self.mock_bg.alleles[AlleleState.CO] = [pair_to_remove]

#         filter_co_existing_pairs({1: self.mock_bg})

#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_to_remove], "filter_co_existing_pairs", AlleleState.CO
#         )

#     def test_filter_co_existing_in_other_allele(self):
#         """Test removal of a 'mushed' pair that matches a different existing allele's variants."""
#         mushed_variants = {"varA", "varB"}
#         allele_mushed = MockAllele("A+B", mushed_variants, genotype_alt="mushed")
#         allele_real = MockAllele(
#             "C_real", mushed_variants
#         )  # Same variants, different genotype
#         allele_other = MockAllele("D", {"varD"})
#         pair_to_remove = MockPair(allele_mushed, allele_other)

#         self.mock_bg.alleles[AlleleState.CO] = [pair_to_remove]
#         self.mock_bg.alleles[AlleleState.FILT] = [allele_real]

#         filter_co_existing_in_other_allele({1: self.mock_bg})

#         self.mock_bg.remove_pairs.assert_called_once_with(
#             [pair_to_remove], "filter_co_existing_in_other_allele", AlleleState.CO
#         )


# class TestEnsureCoExistingHetSnpUsed(unittest.TestCase):
#     def test_no_co_alleles(self):
#         """Test when bg.alleles[AlleleState.CO] is None."""
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: None},
#             sample="sample1",
#             variant_pool={"variant1": Zygosity.HET},
#             filtered_out=defaultdict(list),
#         )
#         result_bg = list(ensure_co_existing_HET_SNP_used({1: bg}).values())[0]
#         self.assertIs(result_bg, bg)
#         self.assertIsNone(result_bg.alleles[AlleleState.CO])

#     def test_no_het_variants(self):
#         """Test when there are no heterozygous variants in variant_pool."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair]},
#             sample="sample2",
#             variant_pool={"variant3": Zygosity.HOM},
#             filtered_out=defaultdict(list),
#         )
#         result_bg = list(ensure_co_existing_HET_SNP_used({1: bg}).values())[0]
#         self.assertIs(result_bg, bg)
#         self.assertIn(pair, result_bg.alleles[AlleleState.CO])
#         self.assertNotIn(pair, result_bg.filtered_out["ensure_HET_SNP_used_CO"])

#     def test_het_variants_no_matching_alleles(self):
#         """Test when there are heterozygous variants but no matching alleles are found."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         allele3 = Allele(
#             genotype="Allele3",
#             phenotype="Phenotype3",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var3"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair_other = Pair(allele1=allele3, allele2=allele3)
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair, pair_other]},
#             sample="sample3",
#             variant_pool={"variant1": Zygosity.HET},
#             filtered_out=defaultdict(list),
#         )
#         result_bg = list(ensure_co_existing_HET_SNP_used({1: bg}).values())[0]
#         self.assertIs(result_bg, bg)
#         self.assertIn(pair, result_bg.alleles[AlleleState.CO])
#         self.assertIn(pair_other, result_bg.alleles[AlleleState.CO])
#         self.assertNotIn(pair, result_bg.filtered_out["ensure_HET_SNP_used_CO"])

#     def test_hits_not_greater_than_one(self):
#         """Test when hits are not greater than 1, pairs are not removed."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         allele3 = Allele(
#             genotype="Allele3",
#             phenotype="Phenotype3",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1", "variant1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair_other = Pair(allele1=allele3, allele2=allele3)
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair, pair_other]},
#             sample="sample4",
#             variant_pool={"variant1": Zygosity.HET},
#             filtered_out=defaultdict(list),
#         )
#         result_bg = list(ensure_co_existing_HET_SNP_used({1: bg}).values())[0]
#         self.assertIs(result_bg, bg)
#         self.assertIn(pair, result_bg.alleles[AlleleState.CO])
#         self.assertNotIn(pair, result_bg.filtered_out["ensure_HET_SNP_used_CO"])

#     def test_hits_greater_than_one_pair_removed(self):
#         """Test when hits > 1, pairs are removed."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         allele3 = Allele(
#             genotype="Allele3",
#             phenotype="Phenotype3",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1", "variant1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele4 = Allele(
#             genotype="Allele4",
#             phenotype="Phenotype4",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2", "variant1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair_other1 = Pair(allele1=allele3, allele2=allele3)
#         pair_other2 = Pair(allele1=allele4, allele2=allele4)
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair, pair_other1, pair_other2]},
#             sample="sample5",
#             variant_pool={"variant1": Zygosity.HET},
#             filtered_out=defaultdict(list),
#         )
#         result_bg = list(ensure_co_existing_HET_SNP_used({1: bg}).values())[0]
#         self.assertIs(result_bg, bg)
#         self.assertNotIn(pair, result_bg.alleles[AlleleState.CO])
#         self.assertIn(pair, result_bg.filtered_out["ensure_co_existing_HET_SNP_used"])
#         self.assertIn(pair_other1, result_bg.alleles[AlleleState.CO])
#         self.assertIn(pair_other2, result_bg.alleles[AlleleState.CO])


# class TestFilterCoExistingInOtherAllele(unittest.TestCase):
#     def test_pairs_removed(self):
#         allele1 = Allele(
#             genotype="FUT2*01.03.01",
#             phenotype="Type1",
#             genotype_alt="mushed",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         allele2 = Allele(
#             genotype="FUT2*01.06",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=2,
#             reference=False,
#             sub_type="subtype2",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         raw_allele = Allele(
#             genotype="FUT2*01.03.03",
#             phenotype="Type3",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=3,
#             reference=False,
#             sub_type="subtype3",
#         )
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair], AlleleState.FILT: [raw_allele]},
#             sample="sample1",
#             variant_pool={},
#         )
#         bg_filtered = list(filter_co_existing_in_other_allele({1: bg}).values())[0]
#         self.assertFalse(bg_filtered.alleles[AlleleState.CO])

#     def test_no_removal_required(self):
#         allele1 = Allele(
#             genotype="FUT2*01.03.01",
#             phenotype="Type1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         allele2 = Allele(
#             genotype="FUT2*01.06",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var3"}),
#             null=False,
#             weight_geno=2,
#             reference=False,
#             sub_type="subtype2",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         raw_allele = Allele(
#             genotype="FUT2*01.03.03",
#             phenotype="Type3",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=3,
#             reference=False,
#             sub_type="subtype3",
#         )
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair], "raw": [raw_allele]},
#             sample="sample1",
#             variant_pool={},
#         )

#         bg_filtered = list(filter_co_existing_in_other_allele({1: bg}).values())[0]
#         self.assertTrue(bg_filtered.alleles[AlleleState.CO])

#     def test_empty_co_existing_list(self):
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [], "raw": []},
#             sample="sample1",
#             variant_pool={},
#         )

#         bg_filtered = list(filter_co_existing_in_other_allele({1: bg}).values())[0]
#         self.assertEqual(bg_filtered.alleles[AlleleState.CO], [])


# class TestFilterCoExistingPairs(unittest.TestCase):
#     def test_normal_case(self):
#         allele1 = Allele(
#             genotype="A1",
#             phenotype="M",
#             genotype_alt="mushed",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="A/B",
#         )
#         allele2 = Allele(
#             genotype="A2",
#             phenotype="N",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=2,
#             reference=False,
#             sub_type="C/D",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair]},
#             sample="sample1",
#             variant_pool={},
#         )

#         bg_filtered = list(filter_co_existing_pairs({1: bg}).values())[0]
#         self.assertFalse(bg_filtered.alleles[AlleleState.CO])

#     def test_no_removal_required(self):
#         allele1 = Allele(
#             genotype="A1",
#             phenotype="M",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="A/B",
#         )
#         allele2 = Allele(
#             genotype="A2",
#             phenotype="N",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=2,
#             reference=False,
#             sub_type="C/D",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair]},
#             sample="sample1",
#             variant_pool={},
#         )

#         bg_filtered = list(filter_co_existing_pairs({1: bg}).values())[0]
#         self.assertTrue(bg_filtered.alleles[AlleleState.CO])

#     def test_empty_co_existing_list(self):
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: []},
#             sample="sample1",
#             variant_pool={},
#         )

#         bg_filtered = list(filter_co_existing_pairs({1: bg}).values())[0]
#         self.assertEqual(bg_filtered.alleles[AlleleState.CO], [])


# class TestFilterCoExistingSubsets(unittest.TestCase):
#     def setUp(self):
#         self.allele1 = Allele(
#             genotype="KN*01.06",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(
#                 {"207782856_A_G, 207782916_A_T, 207782889_A_G, 207782931_A_G"}
#             ),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         self.allele2 = Allele(
#             genotype="KN*01.07",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"207782916_A_T, 207782889_A_G"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         self.allele3 = Allele(
#             genotype="KN*01.10",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({" 207782916_A_T, 207782931_A_G"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         self.allele4 = Allele(
#             genotype="KN*01.07+KN*01.10",
#             phenotype="Type2",
#             genotype_alt="mushed",
#             phenotype_alt=".",
#             defining_variants=frozenset(
#                 {"207782916_A_T, 207782889_A_G, 207782931_A_G"}
#             ),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         self.pair1 = Pair(allele1=self.allele1, allele2=self.allele4)  # Not a subset
#         self.pair2 = Pair(allele1=self.allele4, allele2=self.allele4)  # Subset

#     def test_pairs_removed(self):
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [self.pair1, self.pair2], AlleleState.RAW: []},
#             sample="sample230",
#             variant_pool={
#                 "1:207782856_A_G": "Heterozygous",
#                 "1:207782889_A_G": "Homozygous",
#                 "1:207782916_A_T": "Homozygous",
#                 "1:207782931_A_G": "Homozygous",
#             },
#             filtered_out=defaultdict(list),
#         )
#         self.assertTrue(self.pair2 in bg.alleles[AlleleState.CO])
#         filtered_bg = list(filter_co_existing_subsets({1: bg}).values())[0]

#         self.assertTrue(self.pair2 not in filtered_bg.alleles[AlleleState.CO])
#         self.assertTrue(
#             self.pair2 in filtered_bg.filtered_out["filter_co_existing_subsets"]
#         )

#     def test_no_pairs_removed(self):
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [self.pair1, self.pair2], AlleleState.RAW: []},
#             sample="sample230",
#             variant_pool={},
#             filtered_out=defaultdict(list),
#         )
#         # Reverse the condition so no subsets exist
#         filtered_bg = list(filter_co_existing_subsets({1: bg}).values())[0]
#         self.assertTrue(self.pair1 in filtered_bg.alleles[AlleleState.CO])
#         self.assertTrue(
#             self.pair1 not in filtered_bg.filtered_out["filter_co_existing_subsets"]
#         )  ###TODO wtf - is 'in' not working for Pair objects?

#     def test_empty_co_existing_list(self):
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: []},
#             sample="sample230",
#             variant_pool={},
#             filtered_out=defaultdict(list),
#         )
#         filtered_bg = list(filter_co_existing_subsets({1: bg}).values())[0]
#         self.assertEqual(filtered_bg.alleles[AlleleState.CO], [])

#     def test_permutation_pairs(self):
#         # Create Allele objects
#         alleleA = Allele(
#             genotype="AlleleA",
#             phenotype="TypeA",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"varA"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         alleleB = Allele(
#             genotype="AlleleB",
#             phenotype="TypeB",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"varB"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )

#         # Create pairs that are permutations of each other
#         pair1 = Pair(allele1=alleleA, allele2=alleleB)
#         pair2 = Pair(allele1=alleleB, allele2=alleleA)

#         # Prepare the BloodGroup object
#         bg = BloodGroup(
#             type="KN",
#             alleles={
#                 AlleleState.CO: [pair1, pair2],
#                 AlleleState.RAW: [alleleA, alleleB],
#             },
#             sample="sample231",
#             variant_pool={"varA": Zygosity.HOM, "varB": Zygosity.HOM},
#             filtered_out=defaultdict(list),
#         )

#         # Now run the function
#         filtered_bg = list(filter_co_existing_subsets({1: bg}).values())[0]

#         # Both pairs should remain because they are permutations
#         self.assertIn(pair1, filtered_bg.alleles[AlleleState.CO])
#         self.assertIn(pair2, filtered_bg.alleles[AlleleState.CO])

#     def test_alleles_with_one_het_variant_not_same_subtype(self):
#         # Create Allele objects
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Type1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2", "var3"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype2",  # Different subtype
#         )

#         pair = Pair(allele1=allele1, allele2=allele2)

#         # Prepare the BloodGroup object
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: [pair], AlleleState.RAW: [allele1, allele2]},
#             sample="sample232",
#             variant_pool={
#                 "var1": Zygosity.HET,
#                 "var2": Zygosity.HOM,
#                 "var3": Zygosity.HOM,
#             },
#             filtered_out=defaultdict(list),
#         )

#         # Run the function
#         filtered_bg = list(filter_co_existing_subsets({1: bg}).values())[0]

#         # The pair should not be removed due to the 'continue' in the branch
#         self.assertIn(pair, filtered_bg.alleles[AlleleState.CO])
#         self.assertNotIn(pair, filtered_bg.filtered_out["filter_co_existing_subsets"])

#     def test_contains_reference_pair_removed(self):
#         # Create Reference Allele
#         ref_allele = Allele(
#             genotype="RefAllele",
#             phenotype="RefType",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"ref_var"}),
#             null=False,
#             weight_geno=1,
#             reference=True,  # Mark as reference
#             sub_type="subtype1",
#         )

#         # Create Non-Reference Alleles
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Type1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )

#         # Create Combined Allele
#         combined_allele = Allele(
#             genotype="RefAllele+Allele1",
#             phenotype="CombinedType",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"ref_var", "var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )

#         # Create Pairs
#         pair_with_ref = Pair(allele1=ref_allele, allele2=allele1)
#         other_pair = Pair(allele1=combined_allele, allele2=allele2)

#         # Prepare the BloodGroup object
#         bg = BloodGroup(
#             type="KN",
#             alleles={
#                 AlleleState.CO: [pair_with_ref, other_pair],
#                 AlleleState.RAW: [ref_allele, allele1, allele2, combined_allele],
#             },
#             sample="sample233",
#             variant_pool={
#                 "ref_var": Zygosity.HOM,
#                 "var1": Zygosity.HOM,
#                 "var2": Zygosity.HOM,
#             },
#             filtered_out=defaultdict(list),
#         )

#         # Run the function
#         filtered_bg = list(filter_co_existing_subsets({1: bg}).values())[0]

#         # Assert that pair_with_ref is removed
#         self.assertNotIn(pair_with_ref, filtered_bg.alleles[AlleleState.CO])
#         self.assertIn(
#             pair_with_ref, filtered_bg.filtered_out["filter_co_existing_subsets"]
#         )

#         # Assert that other_pair remains
#         self.assertIn(other_pair, filtered_bg.alleles[AlleleState.CO])


# class TestFilterCoExistingWithNormal(unittest.TestCase):
#     def setUp(self):
#         self.allele1 = Allele(
#             genotype="KN1",
#             phenotype="Type1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )

#         self.allele2 = Allele(
#             genotype="KN2",
#             phenotype="Type2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=2,
#             reference=False,
#             sub_type="subtype2",
#         )
#         # Normal pair
#         self.pair1 = Pair(allele1=self.allele1, allele2=self.allele2)
#         self.allele3 = Allele(
#             genotype="KN1+KN2",
#             phenotype="Type2",
#             genotype_alt="mushed",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1, var2"}),
#             null=False,
#             weight_geno=2,
#             reference=False,
#             sub_type="subtype2",
#         )
#         # Pair with coexisting alleles
#         self.pair2 = Pair(allele1=self.allele1, allele2=self.allele3)

#     def test_normal_case(self):
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.NORMAL: [self.pair1], AlleleState.CO: [self.pair1]},
#             sample="sample230",
#             variant_pool={},
#             filtered_out=defaultdict(list),
#         )
#         self.assertTrue(self.pair1 in bg.alleles[AlleleState.NORMAL])
#         self.assertTrue(self.pair1 in bg.alleles[AlleleState.CO])
#         bg.remove_pairs([self.pair1], "normal_filter", AlleleState.NORMAL)
#         self.assertTrue(self.pair1 not in bg.alleles[AlleleState.NORMAL])
#         filtered_bg = list(filter_co_existing_with_normal({1: bg}).values())[0]
#         self.assertTrue(
#             self.pair1 in filtered_bg.filtered_out["filter_co_existing_with_normal"]
#         )
#         self.assertEqual(filtered_bg.alleles, {"co_existing": [], "pairs": []})

#     def test_no_removal_needed(self):
#         # Modify pair2 to meet exclusion criteria
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.NORMAL: [self.pair1], AlleleState.CO: [self.pair2]},
#             sample="sample230",
#             variant_pool={},
#             filtered_out=defaultdict(list),
#         )
#         filtered_bg = list(filter_co_existing_with_normal({1: bg}).values())[0]
#         self.assertTrue(self.pair2 in filtered_bg.alleles[AlleleState.CO])
#         self.assertEqual(len(filtered_bg.alleles[AlleleState.CO]), 1)
#         self.assertEqual(len(filtered_bg.alleles[AlleleState.NORMAL]), 1)

#     def test_empty_co_existing_list(self):
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.NORMAL: [self.pair1], AlleleState.CO: []},
#             sample="sample230",
#             variant_pool={},
#             filtered_out=defaultdict(list),
#         )
#         filtered_bg = list(filter_co_existing_with_normal({1: bg}).values())[0]
#         self.assertEqual(filtered_bg.alleles[AlleleState.CO], [])


# class TestFilterCoexistingPairsOnAntitheticalZygosity(unittest.TestCase):
#     def test_no_co_alleles(self):
#         """Test when bg.alleles[AlleleState.CO] is None."""
#         bg = BloodGroup(
#             type="KN",
#             alleles={AlleleState.CO: None},
#             sample="sample1",
#         )
#         antitheticals = {"ExampleType": ["antigen1", "antigen2"]}

#         # result_bg = filter_coexisting_pairs_on_antithetical_zygosity(bg, antitheticals)
#         result_bg = list(
#             filter_coexisting_pairs_on_antithetical_zygosity(
#                 {1: bg}, antitheticals
#             ).values()
#         )[0]

#         self.assertIs(result_bg, bg)
#         self.assertIsNone(result_bg.alleles[AlleleState.CO])

#     def test_bg_type_not_in_antitheticals(self):
#         """Test when bg.type is not in antitheticals."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype2",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)

#         bg = BloodGroup(
#             type="OtherType",
#             alleles={AlleleState.CO: [pair], AlleleState.NORMAL: [allele1, allele2]},
#             sample="sample2",
#         )
#         antitheticals = {"ExampleType": ["antigen1", "antigen2"]}

#         # result_bg = filter_coexisting_pairs_on_antithetical_zygosity(bg, antitheticals)
#         result_bg = list(
#             filter_coexisting_pairs_on_antithetical_zygosity(
#                 {1: bg}, antitheticals
#             ).values()
#         )[0]
#         self.assertIs(result_bg, bg)
#         self.assertIn(pair, result_bg.alleles[AlleleState.CO])

#     def test_flattened_sub_types_length_one(self):
#         """Test when length of flattened_sub_types is 1."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)

#         # For AlleleState.NORMAL, we use Pair objects
#         pair_normal = Pair(allele1=allele1, allele2=allele2)

#         bg = BloodGroup(
#             type="KN",
#             alleles={
#                 AlleleState.CO: [pair],
#                 AlleleState.NORMAL: [pair_normal],  # Now contains Pair objects
#             },
#             sample="sample3",
#         )
#         antitheticals = {"ExampleType": ["antigen1", "antigen2"]}

#         result_bg = list(
#             filter_coexisting_pairs_on_antithetical_zygosity(
#                 {1: bg}, antitheticals
#             ).values()
#         )[0]

#         self.assertIs(result_bg, bg)
#         self.assertIn(pair, result_bg.alleles[AlleleState.CO])

#     def test_pairs_with_flat_sub_types_equal(self):
#         """Test when flat_sub_types == flattened_sub_types."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype2",
#         )
#         pair = Pair(allele1=allele1, allele2=allele2)

#         # For AlleleState.NORMAL, we use Pair objects
#         pair_normal = Pair(allele1=allele1, allele2=allele2)

#         bg = BloodGroup(
#             type="KN",
#             alleles={
#                 AlleleState.CO: [pair],
#                 AlleleState.NORMAL: [pair_normal],  # Now contains Pair objects
#             },
#             sample="sample4",
#         )
#         antitheticals = {"ExampleType": ["antigen1", "antigen2"]}

#         result_bg = list(
#             filter_coexisting_pairs_on_antithetical_zygosity(
#                 {1: bg}, antitheticals
#             ).values()
#         )[0]

#         self.assertIs(result_bg, bg)
#         self.assertIn(pair, result_bg.alleles[AlleleState.CO])

#     def test_pairs_with_flat_sub_types_not_equal(self):
#         """Test when flat_sub_types != flattened_sub_types."""
#         allele1 = Allele(
#             genotype="Allele1",
#             phenotype="Phenotype1",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele2 = Allele(
#             genotype="Allele2",
#             phenotype="Phenotype2",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype1",
#         )
#         allele3 = Allele(
#             genotype="Allele3",
#             phenotype="Phenotype3",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset(),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="Subtype2",
#         )
#         # Create pairs for NORMAL state
#         pair_normal1 = Pair(allele1=allele1, allele2=allele2)
#         pair_normal2 = Pair(allele1=allele1, allele2=allele3)  # Different subtypes

#         pair_co = Pair(allele1=allele1, allele2=allele2)

#         bg = BloodGroup(
#             type="KN",
#             alleles={
#                 AlleleState.CO: [pair_co],
#                 AlleleState.NORMAL: [pair_normal1, pair_normal2],
#             },
#             sample="sample5",
#             filtered_out=defaultdict(list),
#         )
#         antitheticals = {"ExampleType": ["antigen1", "antigen2"]}

#         # Adjusted function call to match the decorator
#         result_bg = list(
#             filter_coexisting_pairs_on_antithetical_zygosity(
#                 {1: bg}, antitheticals
#             ).values()
#         )[0]

#         self.assertIs(result_bg, bg)
#         self.assertNotIn(pair_co, result_bg.alleles[AlleleState.CO])
#         self.assertIn(
#             pair_co, result_bg.filtered_out["filter_co_pairs_on_antithetical_zygosity"]
#         )


# class TestParseBioInfo2(unittest.TestCase):
#     def test_multiple_pairs(self):
#         allele1 = Allele(
#             genotype="A1+A2",
#             phenotype="M",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var1"}),
#             null=False,
#             weight_geno=1,
#             reference=False,
#             sub_type="subtype1",
#         )
#         allele2 = Allele(
#             genotype="B1+B2",
#             phenotype="N",
#             genotype_alt=".",
#             phenotype_alt=".",
#             defining_variants=frozenset({"var2"}),
#             null=False,
#             weight_geno=2,
#             reference=False,
#             sub_type="subtype2",
#         )
#         pair1 = Pair(allele1=allele1, allele2=allele2)
#         pair2 = Pair(allele1=allele2, allele2=allele1)
#         result = parse_bio_info2([pair1, pair2])
#         expected = [
#             [frozenset({"A1", "A2"}), frozenset({"B1", "B2"})],
#             [frozenset({"B1", "B2"}), frozenset({"A1", "A2"})],
#         ]
#         self.assertEqual(result, expected)

#     def test_empty_list(self):
#         result = parse_bio_info2([])
#         expected = []
#         self.assertEqual(result, expected)


# if __name__ == "__main__":
#     unittest.main(argv=["first-arg-is-ignored"], exit=False)
